import abc
import dataclasses as dc
import operator
from functools import wraps
from typing import Any, Dict, Sequence, Callable, Type as PyType, Optional

import statey as st
from statey.syms import func, impl, types, base, utils, api, stack
from statey.syms.object_ import Object


class Method(abc.ABC):
    """
    A method is essentially a factory for creating "computed" attributes
    """

    @abc.abstractmethod
    def bind(self, obj: Object) -> Any:
        """
        Given an input object, return some other object representing this attribute
        """
        raise NotImplementedError

    @abc.abstractmethod
    def return_type(self, obj: Object) -> Any:
        """
        Get the return type of this method for the given object
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class InstanceMethod(Method):
    """
    Instance method implementation
    """

    func: func.Function

    def bind(self, obj: Object) -> Any:
        return utils.bind_function_call(self.func, (obj,), registry=obj._registry)

    def return_type(self, obj: Object) -> Any:
        return self.func.type.return_type


@dc.dataclass(frozen=True)
class CallFunctionMethod(Method):
    """
    Some special logic required for applying functions
    """

    func_type: types.FunctionType

    def bind(self, obj: Object) -> Any:
        """
        Unlike all other methods, the __apply__ method for functions
        """

        def call(*args, **kwargs):
            def apply_method(func: self.func_type) -> obj._type.return_type:
                return utils.wrap_function_call(func, args, kwargs)

            res = obj._inst.map(
                utils.native_function(apply_method, registry=obj._registry)
            )
            return Object(res, frame=stack.frame_snapshot(1))

        return call

    def return_type(self, obj: Object) -> Any:
        return self.func_type.return_type

    @classmethod
    @st.hookimpl
    def get_methods(cls, type: types.Type, registry: "Registry") -> "ObjectMethods":
        if not isinstance(type, types.FunctionType):
            return None
        return ConstantObjectMethods({"__call__": cls(type)})


class ObjectMethods(base.AttributeAccess):
    """
    Accessor interface for object methods
    """

    @abc.abstractmethod
    def get_method(self, name: str) -> Method:
        """
        Get a method by name
        """
        raise NotImplementedError

    def get_attr(self, obj: Object, attr: str) -> Object:
        try:
            method = self.get_method(attr)
        except st.exc.NoSuchMethodError as err:
            raise st.exc.SymbolAttributeError(obj, attr) from err
        return method.bind(obj)


class SimpleObjectMethods(ObjectMethods):
    """
    Simple fetch a dictionary of methods via a method, and use that for the get_method function.
    """

    @abc.abstractmethod
    def methods(self) -> Dict[str, Method]:
        """
        Get all of the methods available for this instance
        """
        raise NotImplementedError

    def get_method(self, name: str) -> Method:
        try:
            return self.methods()[name]
        except KeyError as err:
            raise st.exc.NoSuchMethodError(name) from err


@dc.dataclass(frozen=True)
class ConstantObjectMethods(SimpleObjectMethods):
    """
    ObjectMethods implementation from a dict of methods by name
    """

    method_map: Dict[str, Method]

    def methods(self) -> Dict[str, Method]:
        return self.method_map


NoMethods = ConstantObjectMethods({})


@dc.dataclass(frozen=True)
class CompositeObjectMethods(ObjectMethods):
    """
    Class to combine multiple ObjectMethods instances into one
    """

    object_methods: Sequence[ObjectMethods]

    def get_method(self, name: str) -> Method:
        for methods in self.object_methods:
            try:
                return methods.get_method(name)
            except st.exc.NoSuchMethodError:
                pass
        raise st.exc.NoSuchMethodError(name)


def method_from_function(
    function_obj: func.Function,
    return_type: types.Type = utils.MISSING,
    registry: "Registry" = utils.MISSING,
) -> Method:
    """
    Construct a method given a function. A method a function that is accessed as an
    attribute of an object
    """
    if registry is utils.MISSING:
        registry = st.registry

    instance_arg, *func_args = function_obj.type.args

    if return_type is utils.MISSING:
        return_type = function_obj.type.return_type

    method_type = types.NativeFunctionType(func_args, return_type)

    def bind(self: instance_arg.type) -> method_type:
        def bound(*args, **kwargs):
            arguments = utils.bind_function_args(
                function_obj.type, (self,) + args, kwargs
            )
            return function_obj.apply(arguments)

        return utils.native_function(bound, type=method_type, registry=registry)

    method_function_obj = utils.native_function(bind, registry=registry)
    return InstanceMethod(method_function_obj)


def method(
    func: Callable[[Any], Any] = utils.MISSING,
    return_type: types.Type = utils.MISSING,
    registry: "Registry" = utils.MISSING,
) -> Any:
    """
    method_from_function made into a convenient decorator.
    """

    def dec(_func):
        func_obj = utils.native_function(_func, registry=registry)
        method = method_from_function(
            func_obj, return_type=return_type, registry=registry
        )
        _func._statey_method = method
        return _func

    return dec if func is utils.MISSING else dec(func)


class DeclarativeMethodsMeta(type(ObjectMethods)):
    """
    Collect method objects declared
    """

    def __new__(
        cls, name: str, bases: Sequence[PyType], attrs: Dict[str, Any]
    ) -> PyType:
        super_cls = super().__new__(cls, name, bases, attrs)
        methods = super_cls.__methods__ if hasattr(super_cls, "__methods__") else {}
        methods = methods.copy()

        new_methods = {
            key: val._statey_method
            for key, val in attrs.items()
            if isinstance(getattr(val, "_statey_method", None), Method)
        }
        new_methods.update(
            {key: val for key, val in attrs.items() if isinstance(val, Method)}
        )
        methods.update(new_methods)

        super_cls.__methods__ = methods

        return super_cls


class DeclarativeMethods(ObjectMethods, metaclass=DeclarativeMethodsMeta):
    """
    Simple way to write methods for particular types by simply declaring them
    as instance methods on a DeclarativeMethods subclass
    """

    def get_method(self, name: str) -> Method:
        try:
            return self.__methods__[name]
        except KeyError as err:
            raise st.exc.NoSuchMethodError(name) from err


@dc.dataclass(frozen=True)
class MethodsForType:
    """
    Plugin to map methods to some particular types.Type subclass
    """

    type_cls: PyType[types.Type]
    methods: ObjectMethods

    @st.hookimpl
    def get_methods(self, type: types.Type, registry: "Registry") -> ObjectMethods:
        if not isinstance(type, self.type_cls):
            return None
        return self.methods


@dc.dataclass(frozen=True)
class BinaryMagicMethod(Method):
    """
    Magic methods need to return regular python functions instead of statey objects
    """

    name: str
    instance_type: types.Type
    operation: Callable[[Any, Any], Any]
    return_type_opt: Optional[types.Type] = None

    def bind(self, obj: Object) -> Any:

        return_type = self.return_type(obj)

        def apply_method(other: Any) -> return_type:
            def operate(
                inst: self.instance_type, other: self.instance_type
            ) -> return_type:
                return self.operation(inst, other)

            operate.__name__ = self.name
            out = utils.native_function(operate)(obj, other)
            return Object(out, frame=stack.frame_snapshot(1))

        apply_method.__name__ = self.name
        return apply_method

    def return_type(self, obj: Object) -> Any:
        return self.return_type_opt or self.instance_type


def get_magic_methods():
    return {
        "__eq__": operator.eq,
        "__ne__": operator.ne,
        "__add__": operator.add,
        "__sub__": operator.sub,
        "__mul__": operator.mul,
        "__truediv__": operator.truediv,
        "__floordiv__": operator.floordiv,
        "__pow__": operator.pow,
        "__round__": round,
    }


def get_magic_method_types():
    return {
        "__eq__": types.BooleanType(False),
        "__ne__": types.BooleanType(False),
        "__round__": types.IntegerType(True),
    }


@dc.dataclass(frozen=True)
class BinaryMagicMethods(ObjectMethods):
    """
    Methods that are defined for all statey objects. These act a bit differently than regular
    methods since they need to be called by python internally
    """

    type: types.Type
    method_map: Dict[str, Callable[[Any, Any], Any]] = dc.field(
        default_factory=get_magic_methods
    )
    method_map_type_overrides: Dict[str, types.Type] = dc.field(
        default_factory=get_magic_method_types
    )

    def get_method(self, name: str) -> Method:
        try:
            method = self.method_map[name]
        except KeyError as err:
            raise st.exc.NoSuchMethodError(name) from err
        return BinaryMagicMethod(
            name=name,
            operation=method,
            instance_type=self.type,
            return_type_opt=self.method_map_type_overrides.get(name),
        )

    @classmethod
    @st.hookimpl
    def get_methods(cls, type: types.Type, registry: "Registry") -> ObjectMethods:
        if type.nullable:
            return None
        return cls(type)


@dc.dataclass(frozen=True)
class ExpectedValueMethod(Method):
    """
    Method to easily create expected values
    """

    def bind(self, obj: Object) -> Any:
        def expect(value: Any) -> Object:
            expected_obj = Object(value, obj._type, obj._registry)
            return Object(
                impl.ExpectedValue(obj, expected_obj), frame=stack.frame_snapshot(1)
            )

        return expect

    def return_type(self, obj: Object) -> Any:
        return utils.single_arg_function_type(obj._type, obj._type)


@dc.dataclass(frozen=True)
class BaseObjectMethods(SimpleObjectMethods):
    """
    Defines methods available for all objects
    """

    type: types.Type

    def methods(self) -> Dict[str, Method]:
        return {"__rshift__": ExpectedValueMethod()}

    @classmethod
    @st.hookimpl
    def get_methods(cls, type: types.Type, registry: "Registry") -> ObjectMethods:
        if type.nullable:
            return None
        return cls(type)


@dc.dataclass(frozen=True)
class StringMethods(DeclarativeMethods):
    """
    Methods available for string objects
    """

    @method
    def split(self: str, delim: str) -> Sequence[str]:
        return self.split(delim)

    @method
    def strip(self: str) -> str:
        return self.strip()

    @st.hookimpl
    def get_methods(self, type: types.Type, registry: "Registry") -> ObjectMethods:
        if not isinstance(type, types.StringType) or type.nullable:
            return None
        return self


@dc.dataclass(frozen=True)
class OptionMethod(Method):
    """
    Turn every method call into something like a scala option.map()
    """

    type: types.Type
    base_method: Method

    def return_type(self, obj: Object) -> Any:
        return self.base_method.return_type(obj).with_nullable(True)

    def bind(self, obj: Object) -> Any:

        handler = self.base_method.bind(obj)
        return_type = self.return_type(obj)

        if isinstance(return_type, types.FunctionType):

            nullable_return = return_type.return_type.with_nullable(True)

            def out_handler(*args, **kwargs):
                @utils.native_function
                def option_handler(data: self.type) -> nullable_return:
                    if data is None:
                        return None
                    return handler(*args, **kwargs)

                return option_handler(obj)

            return out_handler

        @utils.native_function
        def option_handler(data: self.type) -> return_type:
            if data is None:
                return None
            return handler

        return option_handler(obj)


@dc.dataclass(frozen=True)
class OptionMethods(ObjectMethods):
    """
    Turn every method call into something like a scala option.map()
    """

    type: types.Type
    methods: ObjectMethods

    def get_method(self, name: str) -> Method:
        base_method = self.methods.get_method(name)
        return OptionMethod(self.type, base_method)

    @classmethod
    @st.hookimpl
    def get_methods(cls, type: types.Type, registry: "Registry") -> ObjectMethods:
        if not type.nullable:
            return None
        non_nullable_methods = registry.get_methods(type.with_nullable(False))
        return cls(type, non_nullable_methods)


OBJECT_METHODS_PLUGINS = [
    CallFunctionMethod,
    BinaryMagicMethods,
    BaseObjectMethods,
    StringMethods(),
    OptionMethods,
]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Replace default object methods classes
    """
    if registry is None:
        registry = st.registry

    for cls in OBJECT_METHODS_PLUGINS:
        registry.register(cls)



from typing import Any, Optional, Sequence, Dict

import marshmallow as ma

from statey.syms import utils


class StateyError(Exception):
    """
    Base class for statey errors
    """


class PlanError(StateyError):
    """
    Statey error during planning
    """


class NullRequired(PlanError):
    """
    Error to raise indicating that we need to recreate the planning given resource
    """


class NotADagError(StateyError):
    """
    Error indicating that some graph is not a DAG
    """

    def __init__(self, graph: "DiGraph", path: Sequence[str]) -> None:
        self.graph = graph
        self.path = path
        super().__init__(
            f'Graph {repr(graph)} is not a DAG! Cycle found: {" -> ".join(path)}.'
        )


class ErrorDuringPlanning(PlanError):
    """
    Generic error wrapping during plan()
    """

    def __init__(self, exception: Exception, tb) -> None:
        self.exception = exception
        self.tb = tb
        super().__init__(
            f"Exception raised during planning: {type(exception).__name__}:"
            f" {exception}."
        )


class ErrorDuringPlanningNode(PlanError):
    """
    Error to indicate that an error was raised while calling a resource's
    plan() method
    """

    def __init__(
        self, path: str, current: Any, config: Any, exception: Exception, tb
    ) -> None:
        self.path = path
        self.current = current
        self.config = config
        self.exception = exception
        self.tb = tb
        super().__init__(
            f"Exception raised during planning {current.state.resource}[{path}]"
            f" {current.state.name} => {config.state.name}: "
            f"{type(exception).__name__}: {exception}"
        )


class SymsError(StateyError):
    """
    Base class for errors in the syms package
    """


class SymsTypeError(SymsError, TypeError):
    """
    TypeError base class for type-related errors
    """


class NotFoundError(SymsTypeError):
    """
    Base class for errors indicating some object can't be found
    """


class NoTypeFound(NotFoundError):
    """
    Error to indicate we could not find a type for some annotation
    """

    def __init__(self, annotation: Any) -> None:
        self.annotation = annotation
        super().__init__(f"No type found for annotation: {annotation}.")


class NoEncoderFound(NotFoundError):
    """
    Error to indicate we could not find an encoder for some type
    """

    def __init__(self, type: "Type", serializable: bool = False) -> None:
        self.type = type
        super().__init__(
            f"No encoder found for type: {type} (serializable={serializable})."
        )


class NoSemanticsFound(NotFoundError):
    """
    Error to indicate we could not find semantics for some type
    """

    def __init__(self, type: "Type") -> None:
        self.type = type
        super().__init__(f"No semantics found for type: {type}.")


class NoResourceFound(NotFoundError):
    """
    Error to indicate no resource could be found for a given name
    """

    def __init__(self, resource_name: str) -> None:
        self.resource_name = resource_name
        super().__init__(f"No resource registered for name: {resource_name}.")


class NoDifferFound(NotFoundError):
    """
    Error to indicate we could not find a differ for some type
    """

    def __init__(self, type: "Type") -> None:
        self.type = type
        super().__init__(f"No differ found for type: {type}.")


class NoObjectFound(NotFoundError):
    """
    Error indicating we could not create an object from an arbitrary value
    """

    def __init__(self, value: Any) -> None:
        self.value = value
        super().__init__(f"Unable to create an object from {value}.")


class NoCasterFound(NotFoundError):
    """
    Error indicating we could not create an object from an arbitrary value
    """

    def __init__(self, from_type: "Type", to_type: "Type") -> None:
        self.from_type = from_type
        self.to_type = to_type
        super().__init__(f"Unable to find caster from {from_type} to {to_type}.")


class NoTypeSerializerFound(NotFoundError):
    """
    Base class for NoTypeSerializerFound errors
    """


class NoTypeSerializerFoundForType(NoTypeSerializerFound):
    """
    Error to indicate no type serializer could be found for a given name
    """

    def __init__(self, type: "Type") -> None:
        self.type = type
        super().__init__(f"No type serializer registered for type: {type}.")


class NoTypeSerializerFoundForData(NoTypeSerializerFound):
    """
    Error to indicate no type serializer could be found for a given name
    """

    def __init__(self, data: Any) -> None:
        self.data = data
        super().__init__(f"No type serializer registered for data: {data}.")


class NoObjectImplementationSerializerFound(NotFoundError):
    """
    Error indicating that an ObjectImplementation serializer couldn't
    be found
    """


class NoObjectImplementationSerializerFoundForImpl(
    NoObjectImplementationSerializerFound
):
    """
    Error for the get_impl_serializer() method
    """

    def __init__(self, impl: "ObjectImplementation", type: "Type") -> None:
        self.impl = impl
        self.type = type
        super().__init__(
            f"No serializer found for object implementation: {impl}, type: {type}"
        )


class NoObjectImplementationSerializerFoundForData(
    NoObjectImplementationSerializerFound
):
    """
    Error for the get_impl_serializer_from_data() method
    """

    def __init__(self, data: Any) -> None:
        self.data = data
        super().__init__(f"No object implementation serializer found for data: {data}")


class NoObjectSerializerFound(NotFoundError):
    """
    Error indicating a serializer could not be found for an object
    """


class NoObjectSerializerFoundForObject(NoObjectSerializerFound):
    """
    Error for the get_object_serializer() method
    """

    def __init__(self, obj: "Object") -> None:
        self.obj = obj
        super().__init__(f"No serializer found for object: {obj}")


class NoObjectSerializerFoundForData(NoObjectSerializerFound):
    """
    Error for the get_object_serializer_from_data() method
    """

    def __init__(self, data: Any) -> None:
        self.data = data
        super().__init__(f"No object serializer found for data: {data}")


class NoSessionSerializerFound(NotFoundError):
    """
    Error indicating a serializer could not be found for a session
    """


class NoSessionSerializerFoundForSession(NoSessionSerializerFound):
    """
    Error for the get_session_serializer() method
    """

    def __init__(self, session: "Session") -> None:
        self.session = session
        super().__init__(f"No serializer found for session: {session}")


class NoSessionSerializerFoundForData(NoSessionSerializerFound):
    """
    Error for the get_session_serializer_from_data() method
    """

    def __init__(self, data: Any) -> None:
        self.data = data
        super().__init__(f"No session serializer found for data: {data}")


class NoNamespaceSerializerFound(NotFoundError):
    """
    Error indicating a serializer could not be found for a namespace
    """


class NoNamespaceSerializerFoundForNamespace(NoNamespaceSerializerFound):
    """
    Error for the get_namespace_serializer() method
    """

    def __init__(self, ns: "Namespace") -> None:
        self.ns = ns
        super().__init__(f"No serializer found for namespace: {ns}")


class NoNamespaceSerializerFoundForData(NoNamespaceSerializerFound):
    """
    Error for the get_namespace_serializer_from_data() method
    """

    def __init__(self, data: Any) -> None:
        self.data = data
        super().__init__(f"No namespace serializer found for data: {data}")


class NoProviderFound(NotFoundError):
    """
    Error indicating we could not get a provider for the given params.
    """

    def __init__(self, name: str, params: Dict[str, Any]) -> None:
        self.name = name
        self.params = params
        super().__init__(
            f"Unable to find provider with name {name} and params {params}."
        )


class NoStateManagerFound(NotFoundError):
    """
    Error indicating we could not get a state manager for the given params.
    """

    def __init__(self) -> None:
        super().__init__("Unable to load a state manager for the current module.")


class NoPluginInstallerFound(NotFoundError):
    """
    Error indicating we could not get a plugin installer for the given name
    """

    def __init__(self, name: str) -> None:
        super().__init__(f"Unable to load a plugin installer for plugin: {name}")


class NamespaceError(SymsError):
    """
    Error raised from a namespace
    """


class DuplicateSymbolKey(NamespaceError):
    """
    Error class indicating that we tried to insert a duplicate key into a session
    """

    def __init__(self, key: str, ns: "Namespace") -> None:
        self.key = key
        self.ns = ns
        super().__init__(
            f'Attempted to insert duplicate key "{key}" into namespace {ns}.'
        )


class SymbolKeyError(NamespaceError, KeyError):
    """
    Error raised by a session to indicate a requested key cannot be resolved
    """

    def __init__(self, key: str, ns: "Namespace") -> None:
        self.key = key
        self.ns = ns
        super().__init__(f'Key "{key}" not found in namespace {ns}.')


class SymbolAttributeError(SymsError, AttributeError):
    """
    Error raised to indicate that an attribute reference cannot be resolved on a symbol
    """

    def __init__(self, value: Any, attr: Any) -> None:
        self.value = value
        self.attr = attr
        super().__init__(f'Could not resolve attribute "{attr}" of value {value}.')


class NoSuchMethodError(SymsError):
    """
    Error to indicate a method could not be found
    """

    def __init__(self, name: str) -> None:
        self.name = name
        super().__init__(f"Unable to find method: {name}")


class FutureError(SymsError):
    """
    impl.Future-related errors
    """


class FutureResultNotSet(FutureError):
    """
    Error indicating that get_result() was called on a future whose
    result was not set yet
    """

    def __init__(self, future: "Future") -> None:
        self.future = future
        super().__init__(f"Result has not yet been set for future: {future}.")


class FutureResultAlreadySet(FutureError):
    """
    Error indicating that get_result() was called on a future whose
    result was not set yet
    """

    def __init__(self, future_or_result: Any) -> None:
        from statey.syms.impl import Future, FutureResult

        self.future_or_result = future_or_result

        if isinstance(future_or_result, FutureResult):
            msg = f"Result has already been set: {future_or_result}."
        elif isinstance(future_or_result, Future):
            future = future_or_result
            msg = f"Result has already been set for future: {future} as {future.get_result()}."
        else:
            msg = f"Result has already been set."

        super().__init__(msg)


class SessionError(SymsError):
    """
    Error raised from a session
    """


class ResolutionError(SessionError):
    """
    Error indicating that some exception was encountered during resolution
    """

    def __init__(self, stack: "ResolutionStack", exception: Exception, tb) -> None:
        self.stack = stack
        self.exception = exception
        self.tb = tb
        obj = self.stack.get_object(self.stack.resolving_obj_id)
        super().__init__(
            f"Encountered exception while resolving {obj}: "
            f"{type(exception).__name__}: {exception}."
        )


class InvalidDataError(SessionError):
    """
    Error indicating that there was an attempt to put invalid data into a session
    """

    def __init__(self, key: str, exception: Exception) -> None:
        self.key = key
        self.exception = exception
        super().__init__(
            f"Encountered error while inserting data into key {key}: "
            f"{type(exception).__name__}: {exception}"
        )


class UnknownError(SessionError):
    """
    Error to short-circuit resolution when an unknown value is encountered
    """

    def __init__(
        self, refs: Sequence["Symbol"] = (), expected: Any = utils.MISSING
    ) -> None:
        self.refs = refs
        self.expected = expected
        super().__init__(f"Unknowns found, refs: {refs}; expected: {expected}")


class MissingDataError(SessionError):
    """
    Error indicating that a symbol cannot be resolved because data is missing
    """

    def __init__(self, key: str, type: "Type", session: "Session") -> None:
        self.key = key
        self.type = type
        self.session = session
        super().__init__(
            f'Key "{key}" has not been set in {session} (type: {type.name}).'
        )


class InputValidationError(ma.ValidationError, SymsError):
    """
    Wrapper for a marshmallow validation error, reraise it as a syms exception
    """


class NonEncodeableTypeError(SymsTypeError):
    """
    Raise to indicate that we tried to encode a non-encodeable value
    """

    def __init__(self, type: "Type") -> None:
        self.type = type
        super().__init__(
            f"Encountered non-serializable type: {type} while attempting to serialize or deserialize."
        )


class ResourceError(StateyError):
    """
    Error indicating some issue with a resource or resource code
    """


class InvalidModificationAction(ResourceError):
    """
    Error indicating that get_action() returned an invalid modification action
    in a SimpleMachine
    """

    def __init__(self, modification_type: "ModificationAction") -> None:
        self.modification_type = modification_type
        super().__init__(
            f"Encountered unhandled modification action {modification_type}."
        )


class ExecutionError(StateyError):
    """
    Error to raise indicating that an execution attempt was not successful
    """

    def __init__(self, exec_info: "ExecutionInfo") -> None:
        from statey.task import TaskStatus

        self.exec_info = exec_info
        tasks_by_status = exec_info.tasks_by_status()

        traces = {}
        for key in tasks_by_status.get(TaskStatus.FAILED, []):
            info = exec_info.task_graph.get_info(key)
            if info.error is not None:
                traces[key] = info.error

        trace_lines = []
        for key, error in traces.items():
            trace_lines.append(f"{key}: {error.format_error_message()}")

        trace_lines.insert(0, "")
        trace_str = "\n".join(trace_lines)

        super().__init__(
            f"Error encountered during execution of a task graph; tasks "
            f"by status: {tasks_by_status}. Errors:{trace_str}"
        )


class ProviderError(StateyError):
    """
    Errors relating to providers.
    """


class ResourceNotFound(ProviderError):
    """
    Error indicating a resource could not be found
    """

    def __init__(self, name: str, provider: "Provider") -> None:
        self.name = name
        self.provider = provider
        super().__init__(f'Provider {provider} could not locate resource "{name}".')


class TaskNotFound(ProviderError):
    """
    Error indicating a task could not be found
    """

    def __init__(self, name: str, provider: "Provider") -> None:
        self.name = name
        self.provider = provider
        super().__init__(f'Provider {provider} could not locate task "{name}".')



import abc
import copy
import dataclasses as dc
from typing import Any, Optional, Callable, Dict, Sequence

import statey as st
from statey.syms import types, utils, Object


class Semantics(abc.ABC):
    """
    Semantics define how we should treat a type within our symbolic API e.g.
    what type does an attribute of an instance of a certain type have?
    """

    type: types.Type

    @abc.abstractmethod
    def attr_semantics(self, attr: Any) -> Optional["Semantics"]:
        """
        Return the semantics of the given attribute of this object, if any. A return
        value of None indicates that `attr` is not an attribute of this object.
        """
        raise NotImplementedError

    def path_semantics(self, path: Sequence[Any]) -> Any:
        """
        Apply attr_semantics() multiple times
        """
        base_semantics = self
        for attr in path:
            base_semantics = base_semantics.attr_semantics(attr)
            if base_semantics is None:
                raise AttributeError(path)
        return base_semantics

    @abc.abstractmethod
    def get_attr(self, value: Any, attr: Any) -> Any:
        """
        Given a value, get the attribute value from an encoded value
        """
        raise NotImplementedError

    @abc.abstractmethod
    def map_objects(self, func: Callable[[Any], Any], value: Any) -> Any:
        """
        Assuming that `value` is value or symbol of this type, apply `func` to it
        (if it is an object) or any child objects otherwise
        """
        raise NotImplementedError

    @abc.abstractmethod
    def clone(self, value: Any) -> Any:
        """
        Return a semantically-correct deep copy of the given value
        """
        raise NotImplementedError

    @abc.abstractmethod
    def expand(self, value: Any) -> Any:
        """
        Given a possibly symbolic, expand it to be a dictionary (possibly containing
        objects)
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class ValueSemantics(Semantics):
    """
    Semantics for regular values--these are the default semantics but structural types offer
    additional functionality.
    """

    type: types.Type

    def attr_semantics(self, attr: Any) -> Optional[Semantics]:
        return None

    def get_attr(self, value: Any, attr: Any) -> Any:
        raise st.exc.SymbolAttributeError(value, attr)

    def map_objects(self, func: Callable[[Any], Any], value: Any) -> Any:
        if isinstance(value, Object):
            return func(value)
        return value

    def expand(self, value: Any) -> Any:
        return value

    @classmethod
    @st.hookimpl
    def get_semantics(cls, type: types.Type, registry: st.Registry) -> Semantics:
        return cls(type)

    def clone(self, value: Any) -> Any:
        if isinstance(value, Object):
            return value.clone()
        return copy.copy(value)


@dc.dataclass(frozen=True)
class ArraySemantics(Semantics):
    """
    Semantics for ArrayTypes
    """

    type: types.Type
    element_semantics: Semantics

    def attr_semantics(self, attr: Any) -> Optional[Semantics]:
        if isinstance(attr, int):
            return self.element_semantics
        if isinstance(attr, slice):
            return self
        el_semantics = self.element_semantics.attr_semantics(attr)
        if el_semantics is not None:
            typ = types.ArrayType(el_semantics.type, self.type.nullable)
            return ArraySemantics(typ, el_semantics)
        return None

    def get_attr(self, value: Any, attr: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, Object):
            return value[attr]
        if isinstance(attr, (int, slice)):
            return value[attr]
        # "exploding"
        out = []
        for item in value:
            out.append(self.element_semantics.get_attr(item, attr))
        return out

    def map_objects(self, func: Callable[[Any], Any], value: Any) -> Any:
        if isinstance(value, Object):
            return func(value)
        if value is None:
            return None
        return [self.element_semantics.map_objects(func, item) for item in value]

    def clone(self, value: Any) -> Any:
        if value is None:
            return None

        if isinstance(value, Object):
            return value.clone()
        return [self.element_semantics.clone(item) for item in value]

    def expand(self, value: Any) -> Any:

        from statey.syms import api

        def func(x):
            if x is None:
                return None
            return [self.element_semantics.expand(val) for val in x]

        return api.map(func, value)

    @classmethod
    @st.hookimpl
    def get_semantics(cls, type: types.Type, registry: st.Registry) -> Semantics:
        if not isinstance(type, types.ArrayType):
            return None
        element_semantics = registry.get_semantics(type.element_type)
        return cls(type, element_semantics)


@dc.dataclass(frozen=True)
class MapSemantics(Semantics):
    """
    Semantics for MapTypes
    """

    type: types.Type
    key_semantics: Semantics
    key_encoder: "Encoder"
    value_semantics: Semantics

    def attr_semantics(self, attr: Any) -> Optional[Semantics]:
        try:
            encoded_attr = self.key_encoder.encode(attr)
        except st.exc.InputValidationError:
            return None
        return self.value_semantics

    def get_attr(self, value: Any, attr: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, Object):
            return value[attr]
        encoded_attr = self.key_encoder.encode(attr)
        try:
            return value[encoded_attr]
        except KeyError:
            return None

    def map_objects(self, func: Callable[[Any], Any], value: Any) -> Any:
        if isinstance(value, Object):
            return func(value)
        if value is None:
            return None
        return {
            self.key_semantics.map_objects(func, key): self.value_semantics.map_objects(
                func, val
            )
            for key, val in value.items()
        }

    def clone(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, Object):
            return value.clone()
        return {
            self.key_semantics.clone(key): self.value_semantics.clone(val)
            for key, val in value.items()
        }

    def expand(self, value: Any) -> Any:

        from statey.syms import api

        def func(x):
            if x is None:
                return None
            return {
                self.key_semantics.expand(key): self.value_semantics.expand(val)
                for key, val in x.items()
            }

        return api.map(func, value)

    @classmethod
    @st.hookimpl
    def get_semantics(cls, type: types.Type, registry: st.Registry) -> Semantics:
        if not isinstance(type, types.MapType):
            return None
        key_semantics = registry.get_semantics(type.key_type)
        key_encoder = registry.get_encoder(type.key_type)
        value_semantics = registry.get_semantics(type.value_type)
        return cls(type, key_semantics, key_encoder, value_semantics)


@dc.dataclass(frozen=True)
class StructSemantics(Semantics):
    """
    Semantics for StructTypes
    """

    type: types.Type
    field_semantics: Dict[str, Semantics]

    def attr_semantics(self, attr: Any) -> Optional[Semantics]:
        if attr in self.field_semantics:
            return self.field_semantics[attr]
        return None

    def get_attr(self, value: Any, attr: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, Object):
            return value[attr]
        return value[attr]

    def map_objects(self, func: Callable[[Any], Any], value: Any) -> Any:
        if isinstance(value, Object):
            return func(value)
        if value is None:
            return None
        return {
            key: semantics.map_objects(func, value[key])
            for key, semantics in self.field_semantics.items()
        }

    def clone(self, value: Any) -> Any:
        if value is None:
            return None

        if isinstance(value, Object):
            return value.clone()
        out = {}
        for key, semantics in self.field_semantics.items():
            out[key] = semantics.clone(value[key])
        return out

    def expand(self, value: Any) -> Any:
        if not self.field_semantics or value is None:
            return value
        # If it is nullable it can't be expanded further
        if isinstance(value, Object) and self.type.nullable:
            return value

        out = {}
        for name, semantics in self.field_semantics.items():
            out[name] = semantics.expand(value[name])
        return out

    @classmethod
    @st.hookimpl
    def get_semantics(cls, type: types.Type, registry: st.Registry) -> Semantics:
        if not isinstance(type, types.StructType):
            return None
        field_semantics = {}
        for field in type.fields:
            field_semantics[field.name] = registry.get_semantics(field.type)
        return cls(type, field_semantics)


# Intentionally a list--this can be mutated if desired
SEMANTICS_CLASSES = [ValueSemantics, ArraySemantics, MapSemantics, StructSemantics]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register semantics plugins
    """
    if registry is None:
        registry = st.registry

    for cls in SEMANTICS_CLASSES:
        registry.register(cls)



"""
Base classes for basic functionality used in several different types of objects
"""
import abc
import dataclasses as dc
import sys
from functools import partial
from typing import Any, Sequence

from statey import exc
from statey.syms.stack import internalcode, rewrite_ctx


class AttributeAccess(abc.ABC):
    """
    Defines some interface for getting attributes for some object
    """

    @abc.abstractmethod
    def get_attr(self, obj: Any, attr: str) -> Any:
        """
        Get some attribute by name on the given object
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class OrderedAttributeAccess(AttributeAccess):
    """
    Combine a set of AttributeAccess instances and attempt to use them in order
    """

    accessors: Sequence[AttributeAccess]

    @internalcode
    def get_attr(self, obj: Any, attr: str) -> Any:
        for accessor in self.accessors:
            try:
                return accessor.get_attr(obj, attr)
            except (exc.SymbolAttributeError, exc.SymbolKeyError):
                pass
        raise exc.SymbolAttributeError(obj, attr)


@dc.dataclass(frozen=True)
class GetattrBasedAttributeAccess(AttributeAccess):
    """
    Simple default python attribute access
    """

    obj: Any

    @internalcode
    def get_attr(self, obj: Any, attr: str) -> Any:
        return partial(getattr(self.obj, attr), obj)


class Proxy(abc.ABC):
    """
    Defines a generic way to bind an attribute access object to some instance
    """

    @property
    def _instance(self) -> Any:
        """
        Allow use of this class
        """
        return self

    @property
    @abc.abstractmethod
    def _accessor(self) -> AttributeAccess:
        """
        Get the AttributeAccess instance for this instance
        """
        raise NotImplementedError

    @internalcode
    def __getitem__(self, attr: Any) -> Any:
        """
        Main API method for accessing attributes. Test all accessors in order.
        """
        return self._accessor.get_attr(self._instance, attr)

    @internalcode
    def __getattr__(self, attr: Any) -> Any:
        try:
            return object.__getattr__(self, attr)
        except AttributeError:
            pass

        return self._accessor.get_attr(self._instance, attr)


@dc.dataclass(frozen=True)
class BoundAttributeAccess(Proxy):
    """
    Simple object to allow attribute access on an arbitrary instance
    """

    instance: dc.InitVar[Any]
    accessor: dc.InitVar[AttributeAccess]
    _instance: Any = dc.field(init=False, default=None)
    _accessor: AttributeAccess = dc.field(init=False, default=None)

    def __post_init__(self, obj: Any, accessor: AttributeAccess) -> None:
        self.__dict__["_instance"] = obj
        self.__dict__["_accessor"] = accessor



import abc
import dataclasses as dc
from typing import Optional, Dict, Any, Sequence

import marshmallow as ma
import pluggy

import statey as st
from statey import exc
from statey.hooks import hookspec, create_plugin_manager, hookimpl
from statey.task import Task


class ProviderIdSchema(ma.Schema):
    """
    Schema for a ProviderId
    """

    name = ma.fields.Str()
    meta = ma.fields.Dict(keys=ma.fields.Str(), values=ma.fields.Str())


@dc.dataclass(frozen=True)
class ProviderId:
    """
    Unique identifier for a provider
    """

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ProviderId":
        data = ProviderIdSchema().load(data)
        return cls(**data)

    name: str
    meta: Dict[str, str] = dc.field(default_factory=dict)

    def __hash__(self) -> int:
        return hash((self.name, frozenset(self.meta.items())))

    def __iter__(self) -> iter:
        return iter((self.name, self.meta))

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert this ProviderId to a dictionary
        """
        return ProviderIdSchema().dump(self)


class Provider(abc.ABC):
    """
    A provider provides methods for accessing resources and tasks.
    """

    id: ProviderId

    @abc.abstractmethod
    def list_resources(self) -> Sequence[str]:
        """
        List the resource names available in this provider
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_resource(self, name: str) -> "Resource":
        """
        Get the resource with the given name.
        """
        raise NotImplementedError

    async def setup(self) -> None:
        """
        Perform any necessary provider setup before running
        tasks.
        """

    async def teardown(self) -> None:
        """
        Perform any necessary provider teardown after it is
        no longer being used.
        """


class ProviderHooks:
    """
    Defines hooks to power the default provider
    """

    @hookspec(firstresult=True)
    def get_resource(self, name: str, provider: Provider) -> "Resource":
        """
        Analogous to the Provider.get_resource() method
        """


def create_default_provider_plugin_manager() -> pluggy.PluginManager:
    """
    Factory to create a fresh DefaultProvider plugin manager
    """
    mgr = create_plugin_manager()
    mgr.add_hookspecs(ProviderHooks)
    return mgr


@dc.dataclass(frozen=True)
class DefaultProviderHooks:
    """
    Default hookimpls for DefaultProvider, separate class to avoid name collisions.
    """

    provider: "DefaultProvider"

    @hookimpl
    def get_resource(self, name: str, provider: Provider) -> "Resource":
        real_name = provider._get_registered_resource_name(name)
        return provider.pm.get_plugin(real_name)

    @hookimpl
    def register(self, plugin: Any, registry: "Registry") -> None:
        if isinstance(plugin, st.Resource):
            self.provider.register(plugin)

    @hookimpl
    def get_provider(
        self, name: str, params: Dict[str, Any], registry: "Registry"
    ) -> Provider:
        if ProviderId(name, params) != self.provider.id:
            return None
        if not registry.pm.is_registered(self):
            registry.register(self)
        return self.provider


@dc.dataclass(frozen=True)
class DefaultProvider(Provider):
    """
    Hook-based
    """

    id: ProviderId = dc.field(default=ProviderId("default"))
    pm: pluggy.PluginManager = dc.field(
        init=False,
        default_factory=create_default_provider_plugin_manager,
        compare=False,
        repr=False,
    )

    def __post_init__(self) -> None:
        self.pm.register(self.hookimpls())
        self.__dict__["_resource_names"] = set()

    def list_resources(self) -> Sequence[str]:
        return sorted(self.__dict__["_resource_names"])

    def get_resource(self, name: str) -> "Resource":
        resource = self.pm.hook.get_resource(name=name, provider=self)
        if resource is None:
            raise exc.ResourceNotFound(name, self)
        return resource

    def _get_registered_resource_name(self, resource_name: str) -> str:
        return f"resource:{resource_name}"

    def register(self, resource: "Resource") -> None:
        self.pm.register(
            resource, name=self._get_registered_resource_name(resource.name)
        )
        self.__dict__["_resource_names"].add(resource.name)

    def hookimpls(self) -> DefaultProviderHooks:
        return DefaultProviderHooks(self)


default_provider = DefaultProvider()


DEFAULT_PROVIDERS = [default_provider]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register default plugins
    """
    if registry is None:
        from statey import registry

    for provider in DEFAULT_PROVIDERS:
        registry.register(provider.hookimpls())



import abc
import dataclasses as dc
from functools import partial
from typing import Any, Dict, Sequence, Iterable, Optional, Callable, Union

import networkx as nx

from statey.syms import base, stack


def get_item_passthrough(name: str) -> Callable[[Any], Any]:
    """
    Get a simple method that will just get an attribute using __getitem__
    """

    def func(self):
        return self[name]

    func.__name__ = name
    return property(func)


@dc.dataclass(frozen=True, repr=False)
class Object(base.Proxy):
    """
    A uniform API for structuring access to statey objects. Should not be
    called directly, use an API method to create objects
    """

    impl: dc.InitVar[Any]
    type: dc.InitVar[Optional[Union["Type", Any]]] = None
    registry: dc.InitVar[Optional["Registry"]] = None
    frame: dc.InitVar[Optional[stack.FrameSnapshot]] = None
    _impl: "ObjectImplementation" = dc.field(init=False, default=None)
    _type: "Type" = dc.field(init=False, default=None)
    _registry: "Registry" = dc.field(init=False, default=None)
    _inst: "Proxy" = dc.field(init=False, default=None)
    _frame: stack.FrameSnapshot = dc.field(init=False, default=None)

    def __class_getitem__(cls, item: Any) -> "Object":
        """
        Create a partial object w/ the given annotation
        """
        return partial(cls, type=item)

    def __post_init__(
        self,
        impl: Any,
        type: Optional["Type"],
        registry: Optional["Registry"],
        frame: Optional[stack.FrameSnapshot],
    ) -> None:
        """
        Set up the Object instance
        """
        import statey as st
        from statey.syms.impl import ObjectImplementation

        if isinstance(impl, Object):
            obj = impl
            impl = obj._impl
            type = type or obj._type
            registry = registry or obj._registry
            frame = frame or obj._frame

        elif not isinstance(impl, ObjectImplementation):
            if registry is None:
                registry = st.registry
            obj = registry.get_object(impl)
            impl = obj._impl
            type = type or obj._type

        if registry is None:
            try:
                registry = impl.registry()
            except NotImplementedError:
                pass

        if registry is None:
            registry = st.registry

        if type is not None and not isinstance(type, st.Type):
            # Convert an annotation to a type
            type = registry.get_type(type)

        if type is None:
            try:
                type = impl.type()
            except NotImplementedError:
                pass

        if type is None:
            raise ValueError(
                f"Object implementation {impl} does not have an implied type, one must "
                f"be passed manually."
            )

        if frame is None:
            try:
                frame = impl.frame()
            except NotImplementedError:
                pass

        if frame is None:
            frame = stack.frame_snapshot(2)

        self.__dict__["_impl"] = impl
        self.__dict__["_type"] = type
        self.__dict__["_registry"] = registry
        self.__dict__["_frame"] = frame

        impl_accessor = base.GetattrBasedAttributeAccess(self._impl)
        self.__dict__["_inst"] = base.BoundAttributeAccess(self, impl_accessor)

    @property
    def _accessor(self) -> base.AttributeAccess:
        methods = self._registry.get_methods(self._type)
        return base.OrderedAttributeAccess((self._impl, methods))

    @property
    def _instance(self) -> Any:
        return self

    def __repr__(self) -> str:
        """
        Produce a string representation of this object
        """
        return self._inst.object_repr()

    @property
    def i(self) -> base.Proxy:
        """
        Convenient access to get an implementation-specific bound attribute accessor
        """
        return base.BoundAttributeAccess(self, self._impl)

    @property
    def m(self) -> base.Proxy:
        """
        Convenient access for object methods
        """
        return base.BoundAttributeAccess(self, self._registry.get_methods(self._type))

    @property
    def _(self) -> Any:
        """
        If this object can be trivially resolved alone, do that here. Otherwise, just
        return self. This is useful to resolve something "as much as possible" without
        having to reason about whether something is an Object vs. a regular Python object
        """
        obj = self
        while isinstance(obj, Object):
            try:
                obj = obj._inst.apply_alone()
            except NotImplementedError:
                return obj
        return obj

    # Magic methods need to be defined on the type itself
    # http://docs.python.org/3/reference/datamodel.html#special-method-lookup
    __call__ = get_item_passthrough("__call__")

    __add__ = get_item_passthrough("__add__")

    __sub__ = get_item_passthrough("__sub__")

    __mul__ = get_item_passthrough("__mul__")

    __floordiv__ = get_item_passthrough("__floordiv__")

    __truediv__ = get_item_passthrough("__truediv__")

    __mod__ = get_item_passthrough("__mod__")

    __pow__ = get_item_passthrough("__pow__")

    __lshift__ = get_item_passthrough("__lshift__")

    __rshift__ = get_item_passthrough("__rshift__")

    __and__ = get_item_passthrough("__and__")

    __xor__ = get_item_passthrough("__xor__")

    __or__ = get_item_passthrough("__or__")

    __lt__ = get_item_passthrough("__lt__")

    __le__ = get_item_passthrough("__le__")

    __eq__ = get_item_passthrough("__eq__")

    __ne__ = get_item_passthrough("__ne__")

    __ge__ = get_item_passthrough("__ge__")

    __gt__ = get_item_passthrough("__gt__")

    __neg__ = get_item_passthrough("__ne__")

    __pos__ = get_item_passthrough("__pos__")

    __abs__ = get_item_passthrough("__abs__")

    __invert__ = get_item_passthrough("__invert__")

    __complex__ = get_item_passthrough("__complex__")

    __int__ = get_item_passthrough("__int__")

    __long__ = get_item_passthrough("__long__")

    __float__ = get_item_passthrough("__float__")

    __oct__ = get_item_passthrough("__oct__")

    __hex__ = get_item_passthrough("__hex__")



try:
    import click
except ImportError as err:
    raise RuntimeError(
        "CLI dependencies are not installed, the CLI will not be usable."
        " The CLI will not be usable. CLI dependencies can be included with the"
        " statey[cli] extra."
    ) from err



import dataclasses as dc
import io
from typing import Optional, Any, Sequence


class RequirementsParsingError(Exception):
    """"""

    def __init__(self, src: Any, line: int, err: Exception) -> None:
        self.src = src
        self.line = line
        self.err = err
        super().__init__(f"Error parsing requirements file, line {line}: {err}")


@dc.dataclass(frozen=True)
class PluginSpec:
    """"""

    name: str
    original: str
    version: Optional[str] = None


def parse_requirement(name: str) -> PluginSpec:
    """
    Parse a requirement name, possibly also including a version. Versions should be
    specified with == e.g. pulumi_aws
    """
    if "==" in name:
        plugin_name, version = name.split("==")
        return PluginSpec(plugin_name, name, version)
    return PluginSpec(name, name)


def parse_requirements(file_or_string: Any) -> Sequence[PluginSpec]:
    """
    Parse a requirements file or sequence of requirements
    """
    input_file = file_or_string

    if isinstance(file_or_string, str):
        input_file = io.StringIO(file_or_string)

    reqs = []

    for idx, line in enumerate(file_or_string):
        if not line or line.lstrip().startswith("#"):
            continue
        try:
            parsed = parse_requirement(line.strip())
        except Exception as err:
            raise RequirementsParsingError(file_or_string, idx + 1, err) from err

        reqs.append(parsed)

    return reqs



import abc
import enum
import dataclasses as dc
from typing import Sequence, Any, Optional, Dict

import pluggy

from statey import create_plugin_manager


UNHASHABLE = object()


class TypeStringToken(enum.Enum):
    """
    Different possible tokens in a type string
    """

    QUESTION_MARK = "?"
    LEFT_BRACE = "["
    RIGHT_BRACE = "]"
    COLON = ":"
    ATTR_NAME = "attr"
    TYPE_NAME = "type"
    COMMA = ","
    TO = "=>"
    LAMBDA = "λ"


class TypeStringRenderer:
    """
    Allow customization of how type strings are rendered
    """

    def render(self, value: str, token: TypeStringToken) -> str:
        return value


def hashable_meta(meta: Dict[str, Any]) -> int:
    """
    Generate a hash for a `meta` dictionary. Note that this will omit non-hashable
    values so it will always succeed, but may not always generate the perfect hash
    """
    items = []
    for key, val in meta.items():
        try:
            hash(val)
        # If the type isn't hashable, we'll just hash the type instead. We
        # can't use the
        except TypeError:
            items.append((key, UNHASHABLE, type(val)))
        else:
            items.append((key, val))

    return frozenset(items)


class Type(abc.ABC):
    """
    A type encapsulates information about
    """

    name: str
    nullable: bool
    meta: Dict[str, Any]

    def render_type_string(self, renderer: Optional[TypeStringRenderer] = None) -> str:
        """
        Render a nice human-readable representation of this type.
        """
        if renderer is None:
            renderer = TypeStringRenderer()
        name_rendered = renderer.render(self.name, TypeStringToken.TYPE_NAME)
        suffix = (
            renderer.render("?", TypeStringToken.QUESTION_MARK) if self.nullable else ""
        )
        return f"{name_rendered}{suffix}"

    @abc.abstractmethod
    def with_nullable(self, nullable: bool) -> "Type":
        """
        Return a copy of this type with the given `nullable` value
        """
        raise NotImplementedError

    @abc.abstractmethod
    def with_meta(self, meta: Dict[str, Any]) -> "Type":
        """
        Return a copy of this type with the given `meta` value
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.render_type_string()

    # API methods to building types more easily
    def __call__(self, **meta: Dict[str, Any]) -> "Type":
        """
        Call method on a type modifies its attributes

        e.g. IntegerType(True)(nullable=False)
        """
        obj = self
        if "nullable" in meta:
            obj = obj.with_nullable(meta.pop("nullable"))
        current = self.meta.copy()
        current.update(meta)
        return obj.with_meta(current)

    def __invert__(self) -> "Type":
        """
        Returns a copy of this type with nullable set to True
        """
        return self.with_nullable(True)

    def __hash__(self) -> int:
        return hash((type(self).__name__, self.nullable, hashable_meta(self.meta)))


class DataClassMixin(abc.ABC):
    """
    Mixin to define with_nullable() and with_meta() by using dc.replace(...)
    """

    def with_nullable(self, nullable: bool) -> Type:
        return dc.replace(self, nullable=nullable)

    def with_meta(self, meta: Dict[str, Any]) -> Type:
        return dc.replace(self, meta=meta)


class ValueType(Type):
    """
    Base class for types that just have a single instance
    """

    nullable: bool
    meta: Dict[str, Any]

    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError


@dc.dataclass(frozen=True, repr=False)
class AnyType(DataClassMixin, Type):
    """
    Type that applies to any value. Not a regular dataclass to avoid issues w/ defaults :(
    """

    nullable: bool = False
    meta: Dict[str, Any] = dc.field(default_factory=dict)

    @property
    def name(self) -> str:
        return "any"

    __hash__ = Type.__hash__


class NumberType(ValueType):
    """
    Base class for numeric types
    """


@dc.dataclass(frozen=True, repr=False)
class IntegerType(DataClassMixin, NumberType):
    nullable: bool = False
    meta: Dict[str, Any] = dc.field(default_factory=dict)
    name: str = dc.field(init=False, repr=False, default="integer")

    __hash__ = Type.__hash__


@dc.dataclass(frozen=True, repr=False)
class FloatType(DataClassMixin, NumberType):
    nullable: bool = False
    meta: Dict[str, Any] = dc.field(default_factory=dict)
    name: str = dc.field(init=False, repr=False, default="float")

    __hash__ = Type.__hash__


@dc.dataclass(frozen=True, repr=False)
class BooleanType(DataClassMixin, ValueType):
    nullable: bool = False
    meta: Dict[str, Any] = dc.field(default_factory=dict)
    name: str = dc.field(init=False, repr=False, default="boolean")

    __hash__ = Type.__hash__


@dc.dataclass(frozen=True, repr=False)
class StringType(DataClassMixin, ValueType):
    nullable: bool = False
    meta: Dict[str, Any] = dc.field(default_factory=dict)
    name: str = dc.field(init=False, repr=False, default="string")

    __hash__ = Type.__hash__


@dc.dataclass(frozen=True, repr=False)
class ArrayType(DataClassMixin, Type):
    """
    An array with some element type
    """

    def __class_getitem__(cls, item: Any) -> Type:
        """
        Create array types
        """
        import statey as st

        typ = st.registry.get_type(item)
        return cls(typ)

    element_type: Type
    nullable: bool = False
    meta: Dict[str, Any] = dc.field(default_factory=dict)

    @property
    def name(self) -> str:
        return "array"

    def render_type_string(self, renderer: Optional[TypeStringRenderer] = None) -> str:
        if renderer is None:
            renderer = TypeStringRenderer()
        type_name = renderer.render(self.name, TypeStringToken.TYPE_NAME)
        element_string = self.element_type.render_type_string(renderer)
        suffix = (
            renderer.render("?", TypeStringToken.QUESTION_MARK) if self.nullable else ""
        )
        lbrace = renderer.render("[", TypeStringToken.LEFT_BRACE)
        rbrace = renderer.render("]", TypeStringToken.RIGHT_BRACE)
        return "".join([type_name, lbrace, element_string, rbrace, suffix])

    def with_element_type(self, element_type: Type) -> Type:
        """
        Return a copy of this type with the given element type
        """
        return dc.replace(self, element_type=element_type)

    def __hash__(self) -> int:
        return hash(
            (
                type(self).__name__,
                self.nullable,
                hashable_meta(self.meta),
                self.element_type,
            )
        )


@dc.dataclass(frozen=True)
class Field:
    """
    Contains information about a single field in a StructType
    """

    name: str
    type: Type


@dc.dataclass(frozen=True, repr=False)
class StructType(DataClassMixin, Type):
    """
    A struct contains an ordered sequence of named fields, any of which
    may or may not be null
    """

    def __class_getitem__(cls, value: Sequence[slice]) -> "Type":
        """
        Ability to use syntax like StructType["a": 1]
        """
        import statey as st

        if not isinstance(value, tuple):
            value = (value,)

        fields = []
        for item in value:

            if not isinstance(item, slice):
                raise TypeError(f"{item} is not a slice.")
            if not isinstance(item.start, str):
                raise TypeError(f"{item.start} is not a string.")
            if item.step is not None:
                raise ValueError(
                    f"{item} contains a non-null step, this is not allowed."
                )

            name = item.start
            typ = st.registry.get_type(item.stop)
            fields.append(Field(name, typ))

        return cls(fields)

    fields: Sequence[Field]
    nullable: bool = False
    meta: Dict[str, Any] = dc.field(default_factory=dict)

    def __post_init__(self) -> None:
        self.__dict__["fields"] = tuple(self.fields)

    @property
    def name(self) -> str:
        return "struct"

    def render_type_string(self, renderer: Optional[TypeStringRenderer] = None) -> str:
        if renderer is None:
            renderer = TypeStringRenderer()
        type_name = renderer.render(self.name, TypeStringToken.TYPE_NAME)
        suffix = (
            renderer.render("?", TypeStringToken.QUESTION_MARK) if self.nullable else ""
        )
        colon = renderer.render(":", TypeStringToken.COLON)
        field_strings = [
            f"{renderer.render(field.name, TypeStringToken.ATTR_NAME)}"
            f"{colon}{field.type.render_type_string(renderer)}"
            for field in self.fields
        ]
        comma_and_space = renderer.render(",", TypeStringToken.COMMA) + " "
        lbrace = renderer.render("[", TypeStringToken.LEFT_BRACE)
        rbrace = renderer.render("]", TypeStringToken.RIGHT_BRACE)
        return "".join(
            [type_name, lbrace, comma_and_space.join(field_strings), rbrace, suffix]
        )

    def __getitem__(self, name: str) -> Field:
        """
        Fetch schema fields by name
        """
        field = next((field for field in self.fields if field.name == name), None)
        if field is None:
            raise KeyError(name)
        return field

    def __contains__(self, name: str) -> bool:
        """
        Indicate whether this StructType has a field with the given name
        """
        return any(field.name == name for field in self.fields)

    def with_fields(self, fields: Sequence[Field]) -> Type:
        """
        Return a copy of this type with the fields replaced
        """
        new_inst = dc.replace(self)
        new_inst.__dict__["fields"] = tuple(fields)
        return new_inst

    def __hash__(self) -> int:
        return hash(
            (
                type(self).__name__,
                self.nullable,
                hashable_meta(self.meta),
                tuple(self.fields),
            )
        )


EmptyType = StructType((), True)


@dc.dataclass(frozen=True, repr=False)
class FunctionType(StructType):
    """
    First class functions :)
    """

    args: Sequence[Field]
    return_type: Type
    nullable: bool = dc.field(init=False, default=False)
    meta: Dict[str, Any] = dc.field(init=False, default_factory=dict)
    fields: Sequence[Field] = dc.field(
        init=False, default=(Field("name", StringType(False)),)
    )

    def with_nullable(self, nullable: bool) -> Type:
        new_inst = dc.replace(self)
        new_inst.__dict__["nullable"] = nullable
        new_inst.__dict__["meta"] = self.meta.copy()
        return new_inst

    def with_meta(self, meta: Dict[str, Any]) -> Type:
        new_inst = dc.replace(self)
        new_inst.__dict__["meta"] = meta
        new_inst.__dict__["nullable"] = self.nullable
        return new_inst

    def render_type_string(self, renderer: Optional[TypeStringRenderer] = None) -> str:
        if renderer is None:
            renderer = TypeStringRenderer()
        type_name = renderer.render(self.name, TypeStringToken.TYPE_NAME)
        suffix = (
            renderer.render(
                TypeStringToken.QUESTION_MARK.value, TypeStringToken.QUESTION_MARK
            )
            if self.nullable
            else ""
        )
        colon = renderer.render(TypeStringToken.COLON.value, TypeStringToken.COLON)

        arg_strings = [
            f"{renderer.render(field.name, TypeStringToken.ATTR_NAME)}"
            f"{colon}{field.type.render_type_string(renderer)}"
            for field in self.args
        ]
        comma_and_space = (
            renderer.render(TypeStringToken.COMMA.value, TypeStringToken.COMMA) + " "
        )
        lbrace = renderer.render(
            TypeStringToken.LEFT_BRACE.value, TypeStringToken.LEFT_BRACE
        )
        rbrace = renderer.render(
            TypeStringToken.RIGHT_BRACE.value, TypeStringToken.RIGHT_BRACE
        )

        args_string = "".join(
            [lbrace, comma_and_space.join(arg_strings), rbrace, suffix]
        )

        return_type_string = self.return_type.render_type_string(renderer)

        to_string = renderer.render(TypeStringToken.TO.value, TypeStringToken.TO)
        lambda_string = renderer.render(
            TypeStringToken.LAMBDA.value, TypeStringToken.LAMBDA
        )

        func_type_string = " ".join([args_string, to_string, return_type_string])
        return "".join([lambda_string, lbrace, func_type_string, rbrace, suffix])

    @property
    def name(self) -> str:
        return "function"

    @property
    def args_type(self) -> Type:
        """
        Get a structtype for the arguments of this function
        """
        return StructType(self.args, all(arg.type.nullable for arg in self.args))

    def __hash__(self) -> int:
        return hash(
            (
                type(self).__name__,
                self.nullable,
                hashable_meta(self.meta),
                tuple(self.args),
                self.return_type,
            )
        )


@dc.dataclass(frozen=True, repr=False)
class NativeFunctionType(FunctionType):
    """
    Regular python implementation of FunctionType
    """

    def __post_init__(self) -> None:
        self.__dict__["fields"] = (
            Field("name", StringType(False)),
            # Serialized function object
            Field("serialized", StringType(False)),
        )

    @property
    def name(self) -> str:
        return "native_function"

    __hash__ = FunctionType.__hash__


@dc.dataclass(frozen=True, repr=False)
class MapType(DataClassMixin, Type):
    """
    An array with some element type
    """

    def __class_getitem__(cls, item: Any) -> Type:
        """
        Create array types
        """
        if not isinstance(item, tuple) or len(item) != 2:
            raise ValueError(f'Expected a tuple of length 2, got "{item}".')
        key_annotation, value_annotation = item

        import statey as st

        key_type = st.registry.get_type(key_annotation)
        value_type = st.registry.get_type(value_annotation)
        return cls(key_type, value_type)

    key_type: Type
    value_type: Type
    nullable: bool = False
    meta: Dict[str, Any] = dc.field(default_factory=dict)

    @property
    def name(self) -> str:
        return "map"

    def render_type_string(self, renderer: Optional[TypeStringRenderer] = None) -> str:
        if renderer is None:
            renderer = TypeStringRenderer()
        type_name = renderer.render(self.name, TypeStringToken.TYPE_NAME)
        key_string = self.key_type.render_type_string(renderer)
        value_string = self.value_type.render_type_string(renderer)
        suffix = (
            renderer.render(
                TypeStringToken.QUESTION_MARK.value, TypeStringToken.QUESTION_MARK
            )
            if self.nullable
            else ""
        )
        lbrace = renderer.render(
            TypeStringToken.LEFT_BRACE.value, TypeStringToken.LEFT_BRACE
        )
        rbrace = renderer.render(
            TypeStringToken.RIGHT_BRACE.value, TypeStringToken.RIGHT_BRACE
        )
        comma = renderer.render(TypeStringToken.COMMA.value, TypeStringToken.COMMA)
        return "".join(
            [type_name, lbrace, key_string, comma, " ", value_string, rbrace, suffix]
        )

    def with_key_type(self, key_type: Type) -> Type:
        """
        Return a copy of this type with the given key type
        """
        return dc.replace(self, key_type=key_type)

    def with_value_type(self, value_type: Type) -> Type:
        """
        Return a copy of this type with the given key type
        """
        return dc.replace(self, value_type=value_type)

    def __hash__(self) -> int:
        return hash(
            (
                type(self).__name__,
                self.nullable,
                hashable_meta(self.meta),
                self.key_type,
                self.value_type,
            )
        )


# Some exported objects
Struct = StructType  # Alias

Array = ArrayType  # Alias

Map = MapType  # Alias

Integer = IntegerType()

Float = FloatType()

Boolean = BooleanType()

Any = AnyType()

String = StringType()



import asyncio
import os
from concurrent.futures import ThreadPoolExecutor
from typing import Sequence, Dict, Any, Optional

import pylumi

import statey as st
from statey.ext.pulumi.constants import PULUMI_ID, PULUMI_NS, PULUMI_HOME_ENV
from statey.ext.pulumi.helpers import (
    parse_provider_schema_response,
    PulumiProviderSchema,
)
from statey.ext.pulumi.resource import PulumiResourceMachine


class PulumiProvider(st.Provider):
    """
    Represents a pulumi provider
    """

    def __init__(
        self,
        id: st.ProviderId,
        schema: PulumiProviderSchema,
        operation_timeout: int = 3600,
    ) -> None:

        self.id = id
        self.schema = schema
        self.operation_timeout = operation_timeout

        self.context = None
        self.pulumi_provider = None
        self.thread_pool = None

        self._resource_cache = {}

    async def setup(self) -> None:
        loop = asyncio.get_running_loop()

        pool = self.thread_pool = ThreadPoolExecutor()
        pool.__enter__()

        # This doesn't seem to be working despite
        # https://github.com/pulumi/pulumi/commit/69743fe2bf261f33db887f0d292622ca6ff0e89c
        # TODO--check on versions
        # if PULUMI_HOME_ENV not in os.environ:
        #     os.environ[PULUMI_HOME_ENV] = ".pulumi"

        ctx = self.context = pylumi.Context()
        provider = self.pulumi_provider = ctx.provider(self.schema.name, self.id.meta)
        # Doesn't work to do this async for some reason
        ctx.setup()
        provider.configure()

    async def teardown(self) -> None:
        loop = asyncio.get_running_loop()

        # Doesn't work w/ async for some reason
        self.context.teardown()

        self.thread_pool.__exit__(None, None, None)
        self.context = self.pulumi_provider = self.thread_pool = None

    def list_resources(self) -> Sequence[str]:
        return sorted(self.schema.resources)

    def get_resource(self, name: str) -> st.Resource:
        if name not in self.schema.resources:
            raise st.exc.NoResourceFound(name)

        resource_schema = self.schema.resources[name]
        # Knowing a resource's ID is necessary for pulumi operation
        output_type = st.struct_add(resource_schema.output_type, (PULUMI_ID, str))
        return self.construct_resource(
            name=name, input_type=resource_schema.input_type, output_type=output_type
        )

    def construct_resource(
        self, name: str, input_type: st.Type, output_type: st.Type
    ) -> st.Resource:
        """
        Construct a new MachineResource for the given resource name and input/output types
        """
        if name in self._resource_cache:
            return self._resource_cache[name]

        machine_cls = type(
            PulumiResourceMachine.__name__,
            (PulumiResourceMachine,),
            {"UP": st.State("UP", input_type, output_type)},
        )

        resource = machine_cls(name, self)
        self._resource_cache[name] = resource

        return resource

    @st.hookimpl
    def get_provider(
        self, name: str, params: Dict[str, Any], registry: st.Registry
    ) -> st.Provider:
        if st.ProviderId(name, params) != self.id:
            return None
        return self

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.id})"


def load_pulumi_provider(
    name: str,
    schema_version: int = 0,
    config: Optional[Dict[str, Any]] = None,
    provider_name: Optional[str] = None,
    operation_timeout: Optional[int] = None,
    register: bool = True,
) -> PulumiProvider:
    """
    Load and register a pulumi provider by name.
    """
    if config is None:
        config = {}

    if provider_name is None:
        provider_name = "pulumi/" + name

    config = config.copy()
    version = config.pop("version", None)

    with pylumi.Context() as ctx, ctx.provider(name, config, version) as provider:
        schema = provider.get_schema(schema_version)

    parsed = parse_provider_schema_response(schema)

    kwargs = {}
    if operation_timeout is not None:
        kwargs["operation_timeout"] = operation_timeout

    provider_obj = PulumiProvider(
        st.ProviderId(provider_name, config), parsed, **kwargs
    )
    if register:
        st.registry.register(provider_obj)

    return provider_obj


class PulumiHooks:
    """
    Default hook implementations for pulumi providers.
    """

    @staticmethod
    @st.hookimpl
    def get_provider(
        name: str, params: Dict[str, Any], registry: st.Registry
    ) -> st.Provider:
        if not name.startswith(PULUMI_NS + "/"):
            return None
        _, provider_name = name.split("/", 1)
        return load_pulumi_provider(provider_name, config=params)


DEFAULT_PLUGINS = [PulumiHooks]


def register(registry: Optional[st.Registry] = None) -> None:
    """
    Register plugins for this module
    """
    if registry is None:
        registry = st.registry

    for plugin in DEFAULT_PLUGINS:
        registry.register(plugin)



import abc
import dataclasses as dc
import math
from itertools import chain
from typing import Tuple, Sequence, Any

from statey.syms import utils


class BasePathParser(abc.ABC):
    """
    Defines the interfact that path parsers must implement
    """

    @abc.abstractmethod
    def validate_name(self, name: str) -> None:
        """
        Validate that the given name can be allocated with this path parser.
        Throw an exception if not.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def split(self, path: str) -> Sequence[Any]:
        """
        Split a path into its native-typed components
        """
        raise NotImplementedError

    @abc.abstractmethod
    def join(self, components: Sequence[Any]) -> str:
        """
        Join the natively-typed components into a path string
        """
        raise NotImplementedError


class PathParser(BasePathParser):
    """
    A path parser joins path components into a string and back again

    Paths like a[:1]/b/cdefa[*]/g

    supports encoding/decoding of utils.EXPLODE, strings, integers and slices
    """

    pathsep: str = "."
    explode_key: str = "*"
    idx_wrap: Tuple[str, str] = ("[", "]")

    @property
    def idx_wrap_left(self) -> str:
        return self.idx_wrap[0]

    @property
    def idx_wrap_right(self) -> str:
        return self.idx_wrap[1]

    def validate_name(self, name: str) -> None:
        for reserved_string in [self.pathsep, *self.idx_wrap]:
            if reserved_string in name:
                raise ValueError(
                    f"Key {name} cannot contain reserved string: ({repr(reserved_string)})."
                )

    def split(self, path: str) -> Sequence[Any]:
        out = []
        # Find first index of pathsep or idx_wrap_left
        try:
            first_sep = path.index(self.pathsep)
        except ValueError:
            first_sep = float("inf")

        try:
            first_idx_wrap = path.index(self.idx_wrap_left)
            # Just checking it exists after the first one
            path.index(self.idx_wrap_right, first_idx_wrap)
        except ValueError:
            first_idx_wrap = float("inf")

        # Nothing left to be parsed
        if not math.isfinite(min(first_sep, first_idx_wrap)):
            return (path,)

        if first_sep < first_idx_wrap:
            start, tail = path.split(self.pathsep, 1)
            return (start,) + self.split(tail)

        pre, post = path.split(self.idx_wrap_left, 1)
        idx_str, post = post.split(self.idx_wrap_right, 1)
        idx_strings = [idx_str]

        # Add any additional index strings
        while post.startswith(self.idx_wrap_left):
            if self.idx_wrap_right not in post:
                break
            idx_str, post = post.split(self.idx_wrap_right, 1)
            idx_str = idx_str[len(self.idx_wrap_left) :]
            idx_strings.append(idx_str)

        if post and not post.startswith(self.pathsep):
            raise ValueError(f'Invalid key: "{path}".')
        # Strip leading pathsep off next component
        post = post[len(self.pathsep) :]

        out = [pre]
        for idx_str in idx_strings:
            # This is a slice
            if idx_str == self.explode_key:
                out.append(utils.EXPLODE)
            elif ":" in idx_str:
                comps = idx_str.split(":")
                if len(comps) == 2:
                    start, stop = comps
                    step = None
                else:
                    start, stop, step = comps

                start, stop, step = map(
                    lambda x: int(x) if x else None, [start, stop, step]
                )
                out.append(slice(start, stop, step))
            # Regular index
            else:
                out.append(int(idx_str))

        if not post:
            return tuple(out)
        return tuple(out) + self.split(post)

    def join(self, components: Sequence[Any]) -> str:
        out = []
        skip = False
        for item, next_item in zip(components, chain(components[1:], [object()])):
            if skip:
                skip = False
                continue
            # For array indices we'll join them here
            if isinstance(next_item, int):
                out.append(
                    "".join(
                        [item, self.idx_wrap_left, str(next_item), self.idx_wrap_right]
                    )
                )
                skip = True
            elif isinstance(next_item, slice):

                def string_value(x):
                    return "" if x is None else str(x)

                if next_item.step is not None:
                    slice_str = ":".join(
                        map(
                            string_value,
                            [next_item.start, next_item.stop, next_item.step],
                        )
                    )
                else:
                    slice_str = (
                        ":".join(map(string_value, [next_item.start, next_item.stop]))
                        or ":"
                    )
                out.append(
                    "".join([item, self.idx_wrap_left, slice_str, self.idx_wrap_right])
                )
                skip = True
            elif next_item is utils.EXPLODE:
                out.append(
                    "".join(
                        [
                            item,
                            self.idx_wrap_left,
                            self.explode_key,
                            self.idx_wrap_right,
                        ]
                    )
                )
                skip = True
            else:
                out.append(item)

        return self.pathsep.join(out)



import abc
import dataclasses as dc
from typing import Any, Optional

import marshmallow as ma
import networkx as nx

import statey as st
from statey.syms import (
    session,
    py_session,
    namespace_serializers as ns_serializers,
    impl,
)


class SessionSerializer(abc.ABC):
    """
    Serializes a session object
    """

    def serialize(self, session: session.Session) -> Any:
        """
        Serialize the session to a JSON-serializable object
        """
        raise NotImplementedError

    def deserialize(self, data: Any) -> session.Session:
        """
        Deserialize the given data to a session object
        """
        raise NotImplementedError


class PythonSessionSchema(ma.Schema):
    """
    Schema for a python session dict
    """

    name = ma.fields.Str(required=True, default="python_session")
    graph = ma.fields.Dict()
    data = ma.fields.Dict()
    keys = ma.fields.Dict(keys=ma.fields.Str())
    ns = ma.fields.Dict()


@dc.dataclass(frozen=True)
class PythonSessionSerializer(SessionSerializer):
    """
    Serializer for a python session
    """

    ns_serializer: ns_serializers.NamespaceSerializer
    registry: "Registry"

    def serialize(self, session: session.Session) -> Any:
        serialized_ns = self.ns_serializer.serialize(session.ns)

        objects = {}
        for key, value in session.data.items():
            if isinstance(value, st.Object):
                objects[key] = value
            else:
                typ = session.ns.resolve(key)
                objects[key] = st.Object(impl.Data(value, typ))

        dag = nx.DiGraph()
        for key, obj in objects.items():
            session._build_symbol_dag(obj, dag)

        serialized_dag = nx.DiGraph()
        for node in dag.nodes:
            obj = dag.nodes[node]["symbol"]
            obj_ser = self.registry.get_object_serializer(obj)
            serialized = obj_ser.serialize(obj)
            serialized_dag.add_node(node, object=serialized)

        serialized_dag.add_edges_from(dag.edges)
        graph_data = nx.to_dict_of_dicts(serialized_dag)

        node_data = {
            node: serialized_dag.nodes[node]["object"] for node in serialized_dag.nodes
        }
        obj_id_map = {key: obj._impl.id for key, obj in objects.items()}

        out = {
            "graph": graph_data,
            "data": node_data,
            "keys": obj_id_map,
            "ns": serialized_ns,
        }
        return PythonSessionSchema().dump(out)

    def deserialize(self, data: Any) -> session.Session:
        loaded_data = PythonSessionSchema().load(data)

        ns = self.ns_serializer.deserialize(loaded_data["ns"])

        graph = nx.from_dict_of_dicts(loaded_data["graph"], create_using=nx.DiGraph)
        for key, value in loaded_data["data"].items():
            graph.nodes[key]["object"] = value

        objects = {}
        session = py_session.PythonSession(ns)

        for node in nx.topological_sort(graph):
            obj_data = graph.nodes[node]["object"]
            obj_ser = self.registry.get_object_serializer_from_data(obj_data)
            obj = obj_ser.deserialize(obj_data, session, objects)
            objects[node] = obj

        for key, obj_id in loaded_data["keys"].items():
            session.set_data(key, objects[obj_id])

        return session

    @classmethod
    @st.hookimpl
    def get_session_serializer(
        cls, session: session.Session, registry: "Registry"
    ) -> SessionSerializer:
        if not isinstance(session, py_session.PythonSession):
            return None
        ns_serializer = registry.get_namespace_serializer(session.ns)
        return cls(ns_serializer, registry)

    @classmethod
    @st.hookimpl
    def get_session_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> SessionSerializer:
        if not isinstance(data, dict) or data.get("name") != "python_session":
            return None
        ns_serializer = registry.get_namespace_serializer_from_data(data["ns"])
        return cls(ns_serializer, registry)


SESSION_SERIALIZER_CLASSES = [PythonSessionSerializer]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register session serializers classes with the given registry
    """
    if registry is None:
        registry = st.registry

    for cls in SESSION_SERIALIZER_CLASSES:
        registry.register(cls)



import abc
from typing import Optional


class PluginInstaller(abc.ABC):
    """
    Implements installation functionality for
    """

    def setup(self) -> None:
        """
        Perform any necessary setup before installing plugins
        """

    def teardown(self) -> None:
        """
        Tear down any resource created in setup()
        """

    @abc.abstractmethod
    def install(self, name: str, version: Optional[str] = None) -> None:
        """
        Install a plugin with the given name and optionally version. Version strings must
        be exact, not ranges.
        """
        raise NotImplementedError



import abc
import dataclasses as dc
import enum
import inspect
import sys
import textwrap as tw
import traceback
from datetime import datetime
from functools import wraps, partial
from typing import (
    Tuple,
    Any,
    Optional,
    Callable,
    Sequence,
    Type as PyType,
    Dict,
    Coroutine,
)

import networkx as nx

import statey as st
from statey.syms import session, types, utils, impl, Object


class TaskStatus(enum.Enum):
    """
    Indicates what stage of execution a given task is in.
    """

    NOT_STARTED = 0
    PENDING = 1
    SKIPPED = 2
    FAILED = 3
    SUCCESS = 4


@dc.dataclass(frozen=True)
class ErrorInfo:
    """
    Container for the for all error information
    """

    exc_type: Optional[PyType[Exception]] = None
    exc_value: Optional[Exception] = None
    exc_tb: Optional["traceback"] = None

    @classmethod
    def exc_info(cls) -> "ErrorInfo":
        """
        Mimics the sys.exc_info() method
        """
        return cls(*sys.exc_info())

    def format_exception(self) -> str:
        """
        Return a formatted version of this exception.
        """
        return "".join(
            traceback.format_exception(self.exc_type, self.exc_value, self.exc_tb)
        )

    def format_error_message(self) -> str:
        """
        Return a single-line error message from this exception.
        """
        return f"{self.exc_type.__name__}: {self.exc_value}"


@dc.dataclass(frozen=True)
class TaskInfo:
    """
    Contains information about the state of a task
    """

    status: TaskStatus
    timestamp: datetime = dc.field(default_factory=datetime.utcnow)
    error: Optional[ErrorInfo] = None
    skipped_by: Optional[str] = None


class Task(abc.ABC):
    """
    Base class for tasks. A task is a unit of computation
    """

    description: Optional[str]

    @abc.abstractmethod
    def always_eager(self) -> bool:
        """
        Indicate that this task should always be executed eagerly, meaning even if
        we are interrupted in the middle of execution if all of its dependencies get
        executed, we will still execute this task as long as we exit cleanly
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_metatask(self) -> bool:
        """
        Indicate whether this task is internal and should not be printed in plans
        by default. If all tasks in a plan are metatasks, it is considered to be
        empty
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def run(self) -> None:
        """
        This should be overridden to implement actual task logic.
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class CoroutineTask(Task):
    """
    Wrap a coroutine in a task
    """

    coro_factory: Callable[[], Coroutine]
    description: Optional[str] = None
    _always_eager: bool = False
    _is_metatask: bool = False

    def always_eager(self) -> bool:
        return self._always_eager

    def is_metatask(self) -> bool:
        return self._is_metatask

    async def run(self) -> None:
        await self.coro_factory()


@dc.dataclass(frozen=True)
class FunctionTaskFactory:
    """
    This is essentially a factory that allows us to implement tasks for a session
    """

    func: Callable[[Any], Any]
    is_always_eager: bool = False
    description: Optional[str] = None

    def __call__(self, *args, **kwargs) -> "SessionTaskSpec":
        """
        Create a SessionTaskSpec from this function and the given arguments
        """
        return SessionTaskSpec(
            func=self.func,
            args=args,
            kwargs=kwargs,
            is_always_eager=self.is_always_eager,
            description=self.description,
        )


def task_wrapper(
    func: Callable[[Any], Any] = utils.MISSING,
    description: Optional[str] = None,
    always_eager: bool = False,
    maybe_instance_method: bool = True,
) -> Callable[[Any], Any]:
    """
    Decorator to wrap `func` as a task factory. `func` should be an asynchronous function.
    """

    def process(_func):
        desc = description
        if desc is None:
            doc_func = _func.func if isinstance(_func, partial) else _func
            doc = getattr(doc_func, "__doc__", None)
            desc = tw.dedent(doc).strip() if doc else desc

        sig = inspect.signature(_func)
        param_names = list(sig.parameters)

        if param_names[:1] == ["self"] and maybe_instance_method:

            @wraps(_func)
            def wrapper(self, *args, **kwargs):
                return FunctionTaskFactory(
                    func=partial(func, self),
                    is_always_eager=always_eager,
                    description=desc,
                )(*args, **kwargs)

        else:

            @wraps(_func)
            def wrapper(*args, **kwargs):
                return FunctionTaskFactory(
                    func=_func, is_always_eager=always_eager, description=desc
                )(*args, **kwargs)

        return wrapper

    if func is utils.MISSING:
        return process
    return process(func)


# Alias for the task() function to allow syntax like st.task.new()
new = task_wrapper


@dc.dataclass(frozen=True)
class SessionTaskSpec(utils.Cloneable):
    """
    A task bound with input data
    """

    func: Callable[[Any], Any]
    args: Sequence[Any]
    kwargs: Dict[str, Any]
    description: Optional[str] = None
    is_always_eager: bool = False
    expected: Any = utils.MISSING

    def _expect(self, value: Any) -> "SessionTaskSpec":
        """
        Set the expectation of the output future for tasks created from this spec.
        """
        return self.clone(expected=value)

    def __rshift__(self, value: Any) -> "SessionTaskSpec":
        return self._expect(value)

    def bind(self, session: session.Session) -> "SessionTask":
        """
        Bind this spec to the given session, returning a SessionTask that can be run independently.
        """
        call_obj = utils.wrap_function_call(
            self.func, self.args, self.kwargs, registry=session.ns.registry
        )
        new_future = impl.Future(tuple(call_obj._impl.arguments.values()))

        return SessionTask(
            session=session,
            func=self.func,
            func_type=call_obj._impl.func.type,
            arguments=call_obj._impl.arguments,
            output_future=new_future,
            output_type=call_obj._type,
            description=self.description,
            is_always_eager=self.is_always_eager,
        )


@dc.dataclass(frozen=True)
class SessionTask(Task):
    """
    A task that can be used symbolically within a session
    """

    session: session.Session
    func: Callable[[Any], Any]
    func_type: types.FunctionType
    arguments: Dict[str, Object]
    output_future: impl.Future
    output_type: types.Type
    description: Optional[str] = None
    is_always_eager: bool = False
    description: Optional[str] = None

    def always_eager(self) -> bool:
        return self.is_always_eager

    def is_metatask(self) -> bool:
        return False

    async def run(self) -> None:
        args = []
        for arg in self.func_type.args:
            ref = self.arguments[arg.name]
            args.append(self.session.resolve(ref))

        result = self.func(*args)

        def handle_result(output, idx=0):
            output_symbol = Object(output, self.output_type, self.session.ns.registry)
            resolved_output = self.session.resolve(output_symbol, decode=False)
            self.output_future.set_result(resolved_output, overwrite=idx > 0)

        # Handle all types of functions:
        # - async generators
        # - async functions
        # - regular generators
        # - regular functions
        if inspect.isasyncgen(result):
            idx = 0
            async for output in result:
                handle_result(output, idx)
                idx += 1
        elif inspect.iscoroutine(result):
            handle_result(await result)
        elif inspect.isgenerator(result):
            for idx, output in enumerate(result):
                handle_result(output, idx)
        else:
            handle_result(result)


@dc.dataclass(frozen=True)
class SessionSwitch(Task):
    """
    A session switch resolves a key in one session, and sets that in another session
    """

    input_session: session.Session
    input_symbol: Object
    output_session: session.Session
    output_key: str
    allow_unknowns: bool = True
    description: Optional[str] = None

    def always_eager(self) -> bool:
        return False

    def is_metatask(self) -> bool:
        return True

    async def run(self) -> None:
        resolved_input = self.input_session.resolve(
            self.input_symbol, allow_unknowns=self.allow_unknowns, decode=False
        )
        self.output_session.set_data(self.output_key, resolved_input)


@dc.dataclass(frozen=True)
class ResourceGraphOperation(Task):
    """
    Defines some operation to perform against a resource graph
    """

    key: str
    resource_graph: "ResourceGraph"

    def always_eager(self) -> bool:
        return True

    def is_metatask(self) -> bool:
        return True


async def async_identity(x):
    """
    Simple async identity function
    """
    return x


@dc.dataclass(frozen=True)
class GraphSetKey(ResourceGraphOperation):
    """
    Set some key in a resource graph
    """

    input_session: session.Session
    input_symbol: Object
    dependencies: Sequence[str] = ()
    remove_dependencies: bool = True
    state: Optional["ResourceState"] = None
    finalize: Callable[[Any], Coroutine] = async_identity
    description: Optional[str] = None

    async def run(self) -> None:
        from statey.resource import StateSnapshot

        data = self.input_session.resolve(self.input_symbol, decode=False)
        state = StateSnapshot(data, self.state)
        final_state = await self.finalize(state)

        self.resource_graph.set(
            key=self.key,
            value=final_state.data,
            type=self.input_symbol._type,
            remove_dependencies=self.remove_dependencies,
            state=final_state.state,
        )
        if self.dependencies:
            self.resource_graph.add_dependencies(self.key, self.dependencies)


@dc.dataclass(frozen=True)
class GraphDeleteKey(ResourceGraphOperation):
    """
    Delete some key in a resource graph.
    """

    input_session: Optional[session.Session] = None
    input_symbol: Optional[Object] = None
    description: Optional[str] = None

    async def run(self) -> None:
        # We resolve this symbol so that we know any upstream
        # tasks that this operation depends on went through successfully
        if self.input_session is not None and self.input_symbol is not None:
            self.input_session.resolve(self.input_symbol)
        self.resource_graph.delete(self.key)


class ResourceGraphOperationSpec(abc.ABC):
    """
    Factory for creating resource graph tasks
    """

    def bind(
        self,
        session: "TaskSession",
        resource_graph: "ResourceGraph",
        checkpoint_key: str,
    ) -> ResourceGraphOperation:
        """
        Bind this spec to a task session and resource graph, producing
        a resource graph task
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class GraphSetKeySpec(ResourceGraphOperationSpec):
    """
    Spec to overwrite a state key
    """

    state: "ResourceState"
    remove_dependencies: bool = False

    def bind(
        self,
        session: "TaskSession",
        resource_graph: "ResourceGraph",
        checkpoint_key: str,
    ) -> ResourceGraphOperation:
        resource = session.ns.registry.get_resource(self.state.resource)
        return GraphSetKey(
            input_session=session,
            input_symbol=self.state.obj,
            remove_dependencies=self.remove_dependencies,
            state=self.state.state,
            key=checkpoint_key,
            resource_graph=resource_graph,
            finalize=resource.finalize,
        )


class TaskSession(session.Session):
    """
    Session subclass that wraps a regular session but handles resources in a special manner.
    """

    def __init__(self, session: session.Session) -> None:
        super().__init__(session.ns)
        self.session = session
        self.checkpoints = {}
        self.tasks = {}
        self.pm.register(self)

    def before_set_checkpoint(
        self, key: str, value: "StateSnapshot"
    ) -> Tuple[Any, types.Type]:
        self.checkpoints[key] = GraphSetKeySpec(value)
        return value.data, value.state.output_type

    def before_set_bind_task(
        self, key: str, value: SessionTaskSpec
    ) -> Tuple[Any, types.Type]:
        self.tasks[key] = bound = value.bind(self)
        out_sym = Object(bound.output_future, bound.output_type, self.ns.registry)
        if value.expected is not utils.MISSING:
            out_sym >>= value.expected
        return out_sym, out_sym._type

    @st.hookimpl
    def before_set(self, key: str, value: Any) -> Tuple[Any, types.Type]:
        if isinstance(value, SessionTaskSpec):
            return self.before_set_bind_task(key, value)
        if isinstance(value, st.StateSnapshot):
            return self.before_set_checkpoint(key, value)
        return None

    def resolve(
        self, symbol: Object, allow_unknowns: bool = False, decode: bool = True
    ) -> Any:
        return self.session.resolve(symbol, allow_unknowns, decode)

    def set_data(self, key: str, data: Any) -> None:
        return self.session.set_data(key, data)

    def delete_data(self, key: str) -> None:
        return self.session.delete_data(key)

    def dependency_graph(self) -> nx.MultiDiGraph:
        return self.session.dependency_graph()

    def task_graph(
        self, resource_graph: "ResourceGraph", checkpoint_key: str
    ) -> nx.DiGraph:
        """
        Creates a graph with only tasks as the nodes. If executed in order
        each task's input will be fully resolved and the entire graph of tasks
        has a deterministic output
        """
        task_subgraph = self.dependency_graph()
        keep = list(self.tasks) + list(self.checkpoints)
        utils.subgraph_retaining_dependencies(task_subgraph, keep)
        for node in task_subgraph.nodes:
            if node in self.tasks:
                task_subgraph.nodes[node]["task"] = self.tasks[node]
            else:
                # Construct checkpoint task
                spec = self.checkpoints[node]
                task = spec.bind(self, resource_graph, checkpoint_key)
                task_subgraph.nodes[node]["task"] = task

        return task_subgraph

    def clone(self) -> "TaskSession":
        new_inst = TaskSession(self.session.clone())
        new_inst.tasks = self.tasks.copy()
        new_inst.checkpoints = self.checkpoints.copy()
        for plugin in self.pm.get_plugins():
            if plugin is not self:
                new_inst.pm.register(plugin)
        return new_inst


def create_task_session(session: Optional[session.Session] = None) -> TaskSession:
    """
    Create a task session with default configuration
    """
    if session is None:
        session = st.create_session()
    return TaskSession(session)



import warnings
from typing import Optional

# Pulumi is an optional dependency, allow this to fail if `pylumi`
# isn't installed.
try:
    from statey.ext import pulumi
except RuntimeError:
    warnings.warn(
        "`pylumi` is not installed, pulumi providers and resources "
        "will not be usable.",
        RuntimeWarning,
    )


def register(
    registry: Optional["Registry"] = None,
    pulumi: Optional[bool] = None,
) -> None:
    """
    Register all default library resources
    """
    if registry is None:
        from statey import registry

    ## Register pulumi plugins only if dependencies are installed
    if pulumi is None:
        try:
            import statey.ext.pulumi
        except RuntimeError:
            pulumi = False
        else:
            pulumi = True

    if pulumi:
        from statey.ext.pulumi import register as register_pulumi

        register_pulumi(registry)



import abc
import base64

import dataclasses as dc
from typing import Type as PyType, Any, Dict, Optional

import marshmallow as ma
import pickle
import pluggy

import statey as st
from statey.syms import types, utils, Object


class Encoder(abc.ABC):
    """
    An encoder encodes data of some with possibly native types info some format
    """

    type: types.Type

    @abc.abstractmethod
    def encode(self, type: types.Type, value: Any) -> Any:
        """
        Given a type and some _non-validated_ value, convert it to a serializable value
        """
        raise NotImplementedError

    @abc.abstractmethod
    def decode(self, type: types.Type, value: Any) -> Any:
        """
        Given a freshly deserialized dictionary, potentially apply some post-processing or wrap
        it in a native type
        """
        raise NotImplementedError


class EncoderHooks:
    """
    Hooks to wrap encoder functionality
    """

    @st.hookspec(firstresult=True)
    def encode(self, value: Any) -> Any:
        """
        Optionally apply some logic to encode the given value. return None if the given value is not handled.
        """

    @st.hookspec(firstresult=True)
    def decode(self, value: Any) -> Any:
        """
        Opposite of the encode() hook
        """


def create_encoder_plugin_manager():
    """
    Factory function to create the default plugin manager for encoders
    """
    pm = st.create_plugin_manager()
    pm.add_hookspecs(EncoderHooks)
    return pm


class HookHandlingEncoder(Encoder):
    """
    Handles hooks properly in encode() and decode()
    """

    def encode(self, value: Any) -> Any:
        result = self.pm.hook.encode(value=value)
        return value if result is None else result

    def decode(self, value: Any) -> Any:
        result = self.pm.hook.decode(value=value)
        return value if result is None else result


@dc.dataclass(frozen=True)
class DefaultEncoder(HookHandlingEncoder, utils.Cloneable):
    """
    The default encoder just handles hooks properly, doesn't do any actual encoding
    """

    type: types.Type
    pm: pluggy.PluginManager = dc.field(
        init=False,
        default_factory=create_encoder_plugin_manager,
        compare=False,
        repr=False,
    )

    @classmethod
    @st.hookimpl
    def get_encoder(
        cls, type: types.Type, registry: "Registry", serializable: bool
    ) -> Encoder:
        """
        The basic encoder behavior just calls hooks, but we should pass through plugins too.
        """
        if serializable:
            return None
        inst = cls(type)
        for plugin in type.meta.get("plugins", []):
            inst.pm.register(plugin)
        return inst


@dc.dataclass(frozen=True)
class MarshmallowEncoder(HookHandlingEncoder):
    """
    Encodeable helper to get all functionality from a field factory
    """

    type: types.Type
    registry: "Registry"
    pm: pluggy.PluginManager = dc.field(
        init=False,
        compare=False,
        repr=False,
        default_factory=create_encoder_plugin_manager,
    )

    @abc.abstractmethod
    def base_marshmallow_field(self, encoding: bool) -> ma.fields.Field:
        """
        Return the marshmallow field for this type
        """
        raise NotImplementedError

    def marshmallow_field(self, encoding: bool) -> ma.fields.Field:
        kws = self._marshmallow_field_kws(self.type.nullable, self.type.meta)
        base = self.base_marshmallow_field(encoding)
        return utils.PossiblySymbolicField(base, self.type, self.registry, **kws)

    def encode(self, data: Any) -> Any:
        # Allow pre-encoding hooks
        data = super().encode(data)
        field = self.marshmallow_field(True)
        with utils.reraise_ma_validation_error():
            # This does the validation
            data = field.deserialize(data)
        # This allows us to leverage marshmallow to do things like encoding
        # dates as strings w/ symmetrical encoding/decoding logic
        return field.serialize("tmp", {"tmp": data})

    def decode(self, data: Any) -> Any:
        with utils.reraise_ma_validation_error():
            value = self.marshmallow_field(False).deserialize(data)
            # Allow post-decoding hooks
            return super().decode(value)

    @staticmethod
    def _marshmallow_field_kws(nullable: bool, meta: Dict[str, Any]) -> Dict[str, Any]:
        default = meta.get("default", utils.MISSING)
        validate = meta.get("validator", utils.MISSING)
        if nullable:
            kws = {
                "required": False,
                "default": None if utils.is_missing(default) else default,
                "missing": None if utils.is_missing(default) else default,
                "allow_none": True,
            }
        elif utils.is_missing(default):
            kws = {"required": True}
        else:
            kws = {"missing": default, "default": default}

        if not utils.is_missing(validate):
            kws["validate"] = validate

        return kws


class MarshmallowValueEncoder(MarshmallowEncoder):
    """
    Simple marshmallow encoder for value types
    """

    base_field: ma.fields.Field
    type_cls: PyType[types.Type]
    serializable: bool

    def base_marshmallow_field(self, encoding: bool) -> ma.fields.Field:
        return self.base_field

    @classmethod
    @st.hookimpl
    def get_encoder(
        cls, type: types.Type, registry: "Registry", serializable: bool
    ) -> Encoder:
        if serializable and not cls.serializable:
            return None
        if isinstance(type, cls.type_cls):
            instance = cls(type, registry)
            for plugin in type.meta.get("plugins", []):
                instance.pm.register(plugin)
            return instance
        return None


@dc.dataclass(frozen=True)
class IntegerEncoder(MarshmallowValueEncoder):
    type_cls = types.IntegerType
    base_field = ma.fields.Int()
    serializable = True


@dc.dataclass(frozen=True)
class FloatEncoder(MarshmallowValueEncoder):
    type_cls = types.FloatType
    base_field = ma.fields.Float()
    serializable = True


@dc.dataclass(frozen=True, repr=False)
class BooleanEncoder(MarshmallowValueEncoder):
    type_cls = types.BooleanType
    base_field = ma.fields.Bool()
    serializable = True


@dc.dataclass(frozen=True, repr=False)
class StringEncoder(MarshmallowValueEncoder):
    type_cls = types.StringType
    base_field = ma.fields.Str()
    serializable = True


@dc.dataclass(frozen=True, repr=False)
class ArrayEncoder(MarshmallowEncoder):
    """
    An array with some element type
    """

    element_encoder: Encoder

    def base_marshmallow_field(self, encoding: bool) -> ma.fields.Field:
        kws = self._marshmallow_field_kws(
            self.element_encoder.type.nullable, self.element_encoder.type.meta
        )
        if encoding:
            kws["serialize"] = lambda x: x
            kws["deserialize"] = self.element_encoder.encode
        else:
            kws["serialize"] = lambda x: x
            kws["deserialize"] = self.element_encoder.decode

        element_field = utils.SingleValueFunction(**kws)
        return ma.fields.List(element_field)

    @classmethod
    @st.hookimpl
    def get_encoder(
        cls, type: types.Type, registry: "Registry", serializable: bool
    ) -> Encoder:
        if not isinstance(type, types.ArrayType):
            return None
        element_encoder = registry.get_encoder(type.element_type, serializable)
        instance = cls(type, registry, element_encoder)
        for plugin in type.meta.get("plugins", []):
            instance.pm.register(plugin)
        return instance


@dc.dataclass(frozen=True, repr=False)
class StructEncoder(MarshmallowEncoder):

    field_encoders: Dict[str, Encoder]

    def base_marshmallow_field(self, encoding: bool) -> ma.fields.Field:
        return ma.fields.Nested(self.marshmallow_schema(encoding))

    def marshmallow_schema(self, encoding: bool) -> ma.Schema:
        fields = {}
        for name, encoder in self.field_encoders.items():
            kws = self._marshmallow_field_kws(encoder.type.nullable, encoder.type.meta)
            if encoding:
                kws["serialize"] = lambda x: x
                kws["deserialize"] = encoder.encode
            else:
                kws["serialize"] = lambda x: x
                kws["deserialize"] = encoder.decode
            fields[name] = utils.SingleValueFunction(**kws)
        return type("StructSchema", (ma.Schema,), fields)()

    @classmethod
    @st.hookimpl
    def get_encoder(
        cls, type: types.Type, registry: "Registry", serializable: bool
    ) -> Encoder:
        if not isinstance(type, types.StructType):
            return None
        encoders = {}
        for field in type.fields:
            encoders[field.name] = registry.get_encoder(field.type, serializable)
        instance = cls(type, registry, encoders)
        for plugin in type.meta.get("plugins", []):
            instance.pm.register(plugin)
        return instance


@dc.dataclass(frozen=True)
class NativeFunctionEncoder(StructEncoder):
    """
    Encoder for native python functions
    """

    module: Any = pickle

    def encode(self, value: Any) -> Any:
        if isinstance(value, Object) or value is None:
            return super().encode(value)

        serialized_bytes = self.module.dumps(value.func)
        converted = {
            "serialized": base64.b64encode(serialized_bytes),
            "name": value.name,
        }
        return super().encode(converted)

    def decode(self, value: Any) -> Any:
        from statey.syms import func

        value = super().decode(value)
        if isinstance(value, Object) or value is None:
            return value
        function_ob = self.module.loads(base64.b64decode(value["serialized"]))
        return func.NativeFunction(self.type, function_ob, value["name"])

    @classmethod
    @st.hookimpl
    def get_encoder(
        cls, type: types.Type, registry: "Registry", serializable: bool
    ) -> Encoder:
        if not isinstance(type, types.NativeFunctionType):
            return None
        as_struct = types.StructType(type.fields, type.nullable, type.meta)
        struct_encoder = registry.get_encoder(as_struct, serializable)
        return cls(type, registry, struct_encoder.field_encoders)


@dc.dataclass(frozen=True, repr=False)
class MapEncoder(MarshmallowEncoder):
    """
    An array with some element type
    """

    key_encoder: Encoder
    value_encoder: Encoder

    def base_marshmallow_field(self, encoding: bool) -> ma.fields.Field:
        key_kws = self._marshmallow_field_kws(
            self.key_encoder.type.nullable, self.key_encoder.type.meta
        )
        if encoding:
            key_kws["serialize"] = lambda x: x
            key_kws["deserialize"] = self.key_encoder.encode
        else:
            key_kws["serialize"] = lambda x: x
            key_kws["deserialize"] = self.key_encoder.decode

        key_field = utils.SingleValueFunction(**key_kws)

        value_kws = self._marshmallow_field_kws(
            self.value_encoder.type.nullable, self.value_encoder.type.meta
        )
        if encoding:
            value_kws["serialize"] = lambda x: x
            value_kws["deserialize"] = self.value_encoder.encode
        else:
            value_kws["serialize"] = lambda x: x
            value_kws["deserialize"] = self.value_encoder.decode

        value_field = utils.SingleValueFunction(**value_kws)
        return ma.fields.Dict(keys=key_field, values=value_field)

    @classmethod
    @st.hookimpl
    def get_encoder(
        cls, type: types.Type, registry: "Registry", serializable: bool
    ) -> Encoder:
        if not isinstance(type, types.MapType):
            return None
        key_encoder = registry.get_encoder(type.key_type, serializable)
        value_encoder = registry.get_encoder(type.value_type, serializable)
        instance = cls(type, registry, key_encoder, value_encoder)
        for plugin in type.meta.get("plugins", []):
            instance.pm.register(plugin)
        return instance


ENCODER_CLASSES = [
    DefaultEncoder,
    IntegerEncoder,
    FloatEncoder,
    BooleanEncoder,
    StringEncoder,
    ArrayEncoder,
    StructEncoder,
    NativeFunctionEncoder,
    MapEncoder,
]


# We'll prefer a better pickling module if we have one.
try:
    import dill
except ImportError:
    import warnings

    warnings.warn("Dill is not installed.", RuntimeWarning)
else:

    @dc.dataclass(frozen=True)
    class DillFunctionEncoder(NativeFunctionEncoder):
        """
        dill-based python function encoder
        """

        module: Any = dill

    ENCODER_CLASSES.append(DillFunctionEncoder)


try:
    import cloudpickle
except ImportError:
    import warnings

    warnings.warn("Cloudpickle is not installed.", RuntimeWarning)
else:

    @dc.dataclass(frozen=True)
    class CloudPickleFunctionEncoder(NativeFunctionEncoder):
        """
        cloudpickle-based python function encoder
        """

        module: Any = cloudpickle

    ENCODER_CLASSES.append(CloudPickleFunctionEncoder)


def register(registry: Optional["Registry"] = None) -> None:
    """
    Replace default encoder with encoders defined here
    """
    if registry is None:
        registry = st.registry

    for cls in ENCODER_CLASSES:
        registry.register(cls)



import warnings

NS = "statey"

__version__ = "0.0.8"

from statey.hooks import (
    hookimpl,
    hookspec,
    create_plugin_manager,
    register_default_plugins,
)

from statey.registry import Registry, create_registry

registry = create_registry()

from statey import (
    exc,
    helpers,
    plan,
    task,
    syms,
)

from statey.fsm import (
    Machine,
    transition,
    SingleStateMachine,
    SimpleMachine,
    ModificationAction,
)

from statey.plan import Plan, PlanAction, Migrator, DefaultMigrator

from statey.plugin_installer import PluginInstaller

from statey.provider import Provider, ProviderId

from statey.syms.diff import Diff, DiffConfig, Differ

# `fmt` is an optional dependency, this import will fail if it isn't installed
try:
    from statey.syms.fmt import f
except RuntimeError:
    warnings.warn(
        "`fmt` is not installed, the st.f() function will not be usable.",
        RuntimeWarning,
    )

from statey.syms.func import Function

from statey.syms.impl import Reference, Unknown

from statey.syms.object_ import Object

from statey.syms.session import Session

from statey.syms.path import PathParser

from statey.syms.py_session import create_session

from statey.syms.types import (
    Type,
    StructType,
    StringType,
    IntegerType,
    FloatType,
    ArrayType,
    MapType,
    BooleanType,
    NumberType,
    Field,
    EmptyType,
    Integer,
    String,
    Float,
    Struct,
    Array,
    Boolean,
    Any,
    Map,
)

from statey.syms.api import (
    F,
    join,
    struct,
    function,
    map,
    declarative,
    struct_replace,
    fill,
    fill_unknowns,
    filter_struct,
    ifnull,
    struct_drop,
    struct_add,
    struct_interpolate,
    str,
    int,
)

from statey.resource import (
    BoundState,
    ResourceSession,
    ResourceState,
    State,
    NullState,
    create_resource_session,
    StateSnapshot,
    StateConfig,
    ResourceGraph,
    Resource,
)

from statey.task import Task, TaskSession, TaskStatus, create_task_session


def set_registry(new_registry: Registry) -> None:
    """
    Set st.registry
    """
    global registry
    registry = new_registry


register_default_plugins(registry)



from typing import Sequence, Dict, Any


class PulumiError(Exception):
    """
    Errors originating from statey.ext.pulumi
    """


class PulumiValidationError(PulumiError):
    """
    Validation error originating from pulumi
    """

    def __init__(self, errors: Sequence[Dict[str, Any]]) -> None:
        self.errors = errors
        super().__init__(f"Parameter validation failed with errors: {errors}.")


class PluginVersionRequired(PulumiError):
    """
    Error indicating that an attempt was made to install a plugin without a version
    """

    def __init__(self, name: str) -> None:
        self.name = name
        super().__init__(f'Cannot install "{name}" without an explicit version.')



import abc
import dataclasses as dc
from functools import partial
from typing import Dict, Any, Optional, Tuple, Union, Iterator

import marshmallow as ma
import networkx as nx

import statey as st
from statey import exc
from statey.syms import types, utils, path, impl, Object, stack


class Namespace(abc.ABC):
    """
    A namespace contains information about names and their associated types
    """

    def __init__(
        self,
        registry: Optional["Registry"] = None,
        path_parser: path.PathParser = path.PathParser(),
    ) -> None:
        if registry is None:
            registry = st.registry
        self.path_parser = path_parser
        self.registry = registry

    @abc.abstractmethod
    def new(self, key: str, type: types.Type) -> Object:
        """
        Create a new symbol for the given key and schema and add it to the current namespace.
        Will raise an error if the key already exists.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, key: str) -> None:
        """
        Delete the type at the given key.

        Note this should be used with EXTREME CAUTION because it can leave dependent sessions
        with invalid references.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def keys(self) -> Iterator[str]:
        """
        Iterate through the currently defined keys in this namespace.
        """
        raise NotImplementedError

    def ref(self, key: str) -> Object:
        """
        Get a reference to a key whose type is already registered in this namespace. Will raise
        SymbolKeyError is none exists
        """
        typ = self.resolve(key)
        semantics = self.registry.get_semantics(typ)
        new_impl = impl.Reference(key, self)
        return Object(
            new_impl, semantics.type, self.registry, frame=stack.frame_snapshot(1)
        )

    @abc.abstractmethod
    def resolve(self, key: str) -> types.Type:
        """
        Get the type of the given key, raising an error if it is not in the current schema
        """
        raise NotImplementedError

    @abc.abstractmethod
    def clone(self) -> "Namespace":
        """
        Return a copy of this namespace
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class NamedSessionSetter:
    """
    This is returned from Session.__getitem__ and can allow a few different
    expression syntax options like
    a = session['a'] << A() # This returns a reference correctly instead of an A object
    """

    key: str
    annotation: Any
    session: "Session"

    def __lshift__(self, other: Any) -> Object:
        ref = self.session.set(self.key, other, self.annotation)
        return Object(ref, frame=stack.frame_snapshot(1))


class SessionHooks:
    """
    Hook for pluggable functionality in sessions
    """

    @st.hookspec(firstresult=True)
    def before_set(self, key: str, value: Any) -> Tuple[Any, types.Type]:
        """
        Hook to handle a value before the usual logic in the set() method. This should
        return a (value, type) tuple, where type can be utils.MISSING to indicate we want
        to continue to use the default type inference logic.
        """

    @st.hookspec(firstresult=True)
    def after_set(self, key: str, value: Any, type: types.Type) -> Any:
        """
        Hook after we've already set a value to customize the value returned instead of a
        reference to the
        """


class Session(abc.ABC):
    """
    A session contains a namespace and associated data and objects
    """

    def __init__(self, ns: Namespace) -> None:
        self.ns = ns
        self.pm = st.create_plugin_manager()
        self.pm.add_hookspecs(SessionHooks)

    def set(self, key: str, value: Any, annotation: Any = utils.MISSING) -> Object:
        """
        Set the given data, using the given registry to determine a schema for value
        """
        hook_resp = self.pm.hook.before_set(key=key, value=value)
        typ = utils.MISSING
        if hook_resp is not None:
            value, typ = hook_resp

        if typ is utils.MISSING:
            if isinstance(value, Object):
                typ = value._type
            elif annotation is utils.MISSING:
                typ = self.ns.registry.infer_type(value)
            else:
                typ = self.ns.registry.get_type(annotation)

        ref = self.ns.new(key, typ)
        ref = Object(ref, frame=stack.frame_snapshot(1))
        self.set_data(key, value)

        other_result = self.pm.hook.after_set(key=key, value=value, type=typ)
        return ref if other_result is None else other_result

    def delete(self, key: str) -> None:
        """
        Convenience method to delete the key from the namespace, then delete the data
        """
        self.ns.delete(key)
        self.delete_data(key)

    def symbolify(self, data: Any, type: types.Type = utils.MISSING) -> Object:
        """
        Convert the input data into a symbol, optionally with the given type.
        If it is already a symbol, it will be returned unchanged.
        """
        if isinstance(data, Object):
            return data

        if type is utils.MISSING:
            type = self.ns.registry.infer_type(data)
        semantics = self.ns.registry.get_semantics(type)
        return Object(impl.Data(data), type, self.ns.registry)

    def __lshift__(self, other: Any) -> Object:
        """
        lshift on the top level of a session will simply convert the input to a symbol
        using symbolify()
        """
        return self.symbolify(other)

    def __setitem__(self, key: Union[slice, str], value: Any) -> None:
        """
        Allow dictionary syntax for adding items to the session
        """
        annotation = utils.MISSING
        if isinstance(key, slice):
            key, annotation = key.start, key.stop
        self.set(key, value, annotation)

    def __getitem__(self, key: Union[slice, str]) -> NamedSessionSetter:
        """
        Return a special object to provide better syntax for certain operations
        """
        annotation = utils.MISSING
        if isinstance(key, slice):
            key, annotation = key.start, key.stop
        return NamedSessionSetter(key, annotation, self)

    # Abstract methods
    @abc.abstractmethod
    def resolve(
        self, symbol: Object, allow_unknowns: bool = False, decode: bool = True
    ) -> Any:
        """
        Resolve the given symbol with the given input data.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_data(self, key: str, data: Any) -> None:
        """
        Set the given data at the given key. Data can be or contain objects provided they
        are correctly typed. `key` must be a ROOT key, not an attribute
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_data(self, key: str) -> None:
        """
        Delete the data at the given key
        """
        raise NotImplementedError

    @abc.abstractmethod
    def dependency_graph(self) -> nx.MultiDiGraph:
        """
        Return a graph whose nodes are the top-level names registered in this session's namespace
        and whose edges are the dependencies between those nodes within this session. Each edge should
        include `path` as a property with the relative path reference that the dependency represents.
        Note that all paths in this result should be returned as tuples so that they are not dependent
        on this session's path parser implementation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def clone(self) -> "Session":
        """
        Return a copy of this session that can be modified without affecting this one
        """
        raise NotImplementedError



from typing import Optional

try:
    import pylumi
except ImportError as err:
    raise RuntimeError(
        f"`pylumi` is not installed, this extension module cannot be used."
    ) from err

from statey.ext.pulumi.plugin_installer import PulumiPluginInstaller
from statey.ext.pulumi.provider import PulumiProvider
from statey.ext.pulumi.resource import PulumiResourceMachine


def register(registry: Optional["Registry"] = None) -> None:
    from . import provider, helpers, plugin_installer

    if registry is None:
        from statey import registry

    helpers.register(registry)
    provider.register(registry)
    plugin_installer.register(registry)



import asyncio
from typing import Dict, Any, Optional

import pylumi

import statey as st
from statey.ext.pulumi.constants import PULUMI_ID
from statey.ext.pulumi.exc import PulumiValidationError
from statey.ext.pulumi.helpers import object_to_pulumi_json


class PulumiResourceMachine(st.SingleStateMachine):
    """"""

    def make_output(
        self, data: Dict[str, Any], id: str, typ: st.Type
    ) -> Dict[str, Any]:
        """"""
        data = dict(data, **{PULUMI_ID: id})
        # Trim any extra fields
        data = {
            field.name: data[field.name] for field in typ.fields if field.name in data
        }
        # Add in default values for any required but missing fields
        for field in typ.fields:
            if field.name in data and data[field.name] is None:
                del data[field.name]

            if field.type.nullable or field.name in data:
                continue

            if isinstance(field.type, st.NumberType):
                data[field.name] = 0
            elif isinstance(field.type, st.StringType):
                data[field.name] = ""
            elif isinstance(field.type, st.BooleanType):
                data[field.name] = False
            elif isinstance(field.type, st.ArrayType):
                data[field.name] = []
            elif isinstance(field.type, (st.MapType, st.StructType)):
                data[field.name] = {}

        encoder = st.registry.get_encoder(typ)
        return encoder.encode(data)

    def make_expected_output(
        self, data: Dict[str, Any], config: st.StateConfig, typ: st.Type
    ) -> st.Object:
        """"""
        stripped = {key: val for key, val in data.items() if val is not None}

        drop_fields = []
        for field in typ.fields:
            if field.name not in stripped:
                drop_fields.append(field.name)

        sub_type = st.struct_drop(typ, *drop_fields)
        data_obj = st.Object(stripped, sub_type)

        unknown = st.Unknown[typ]
        config_type = config.obj._type
        config_fields = {field.name: field for field in config_type.fields}
        registry = config.obj._registry

        def get_value(key):
            this_field = typ[key]
            if key in config_fields:
                config_field = config_fields[key]

                try:
                    caster = registry.get_caster(config_field.type, this_field.type)
                except st.exc.NoCasterFound:
                    pass
                else:
                    return caster.cast(config.obj[key])

                non_nullable_config = config_field.type.with_nullable(False)
                try:
                    caster = registry.get_caster(non_nullable_config, this_field.type)
                except st.exc.NoCasterFound:
                    pass
                else:
                    return caster.cast(st.ifnull(config.obj[key], unknown[key]))

            return unknown[key]

        return st.fill(data_obj, typ, get_value)

    def clean_check_resp(
        self, resp: Dict[str, Any], type: st.Type, registry: st.Registry
    ) -> Dict[str, Any]:
        """
        Clean the check() response
        """
        if isinstance(resp, pylumi.UnknownValue):
            return st.Unknown[type]

        if isinstance(resp, list) and isinstance(type, st.ArrayType):
            return [
                self.clean_check_resp(item, type.element_type, registry)
                for item in resp
            ]

        if not isinstance(resp, dict):
            return resp

        out = {}

        semantics = registry.get_semantics(type)
        for key, val in resp.items():
            if key in {"__defaults"}:
                continue
            val_semantics = semantics.attr_semantics(key)
            out[key] = self.clean_check_resp(val, val_semantics.type, registry)

        return out

    async def get_action(
        self,
        current: st.StateSnapshot,
        config: st.StateConfig,
        session: st.TaskSession,
    ) -> st.ModificationAction:

        loop = asyncio.get_running_loop()

        check_resp, errs = await loop.run_in_executor(
            self.provider.thread_pool,
            self.provider.pulumi_provider.check,
            pylumi.URN(self.name),
            current.data,
            object_to_pulumi_json(config.obj, session),
        )
        check_resp = self.clean_check_resp(
            check_resp, self.UP.input_type, session.ns.registry
        )
        if errs:
            raise PulumiValidationError(errs)

        current_data = current.data.copy()
        pulumi_id = current_data.pop(PULUMI_ID)
        diff_resp = await loop.run_in_executor(
            self.provider.thread_pool,
            self.provider.pulumi_provider.diff,
            pylumi.URN(self.name, pulumi_id),
            pulumi_id,
            current_data,
            check_resp,
        )

        if not diff_resp["DetailedDiff"]:
            return st.ModificationAction.NONE

        replace_kinds = {
            pylumi.DiffKind.ADD_REPLACE,
            pylumi.DiffKind.DELETE_REPLACE,
            pylumi.DiffKind.UPDATE_REPLACE,
        }

        for key, val in diff_resp["DetailedDiff"].items():
            kind = pylumi.DiffKind(val["Kind"])
            if kind in replace_kinds:
                return st.ModificationAction.DELETE_AND_RECREATE

        return st.ModificationAction.MODIFY

    async def get_expected(
        self, current: st.StateSnapshot, config: st.StateConfig, session: st.TaskSession
    ) -> Any:

        loop = asyncio.get_running_loop()

        current_is_up = current.state.name == "UP"
        config_is_up = config.state.name == "UP"

        if not config_is_up:
            return {}

        config_json = object_to_pulumi_json(config.obj, session)
        check_resp, errs = await loop.run_in_executor(
            self.provider.thread_pool,
            self.provider.pulumi_provider.check,
            pylumi.URN(self.name),
            current.data,
            config_json,
            True,
        )
        check_resp = self.clean_check_resp(
            check_resp, self.UP.input_type, session.ns.registry
        )
        check_resp = object_to_pulumi_json(
            st.Object(check_resp, self.UP.input_type), session
        )
        if errs:
            raise PulumiValidationError(errs)

        if not current_is_up:
            resp = await loop.run_in_executor(
                self.provider.thread_pool,
                self.provider.pulumi_provider.create,
                pylumi.URN(self.name),
                check_resp,
                self.provider.operation_timeout,
                True,
            )
            props = self.clean_check_resp(
                resp["Properties"], self.UP.output_type, session.ns.registry
            )
            return self.make_expected_output(props, config, self.UP.output_type)

        current_data = current.data.copy()
        pulumi_id = current_data.pop(PULUMI_ID)
        diff_resp = await loop.run_in_executor(
            self.provider.thread_pool,
            self.provider.pulumi_provider.diff,
            pylumi.URN(self.name, pulumi_id),
            pulumi_id,
            current_data,
            check_resp,
        )

        output = current.data.copy()
        for key in diff_resp["ChangedKeys"]:
            output.pop(key, None)

        return self.make_expected_output(output, config, self.UP.output_type)

    async def refresh_state(self, data: Any) -> Optional[Any]:
        loop = asyncio.get_running_loop()

        data = data.copy()
        pulumi_id = data.pop(PULUMI_ID)
        resp = await loop.run_in_executor(
            self.provider.thread_pool,
            self.provider.pulumi_provider.read,
            pylumi.URN(self.name, pulumi_id),
            pulumi_id,
            data,
            data,
        )
        if not resp["Outputs"]:
            return None
        return self.make_output(resp["Outputs"], pulumi_id, self.UP.output_type)

    async def create(
        self, session: st.TaskSession, config: st.StateConfig
    ) -> st.Object:
        current = st.StateSnapshot({}, self.null_state.state)
        expected = await self.get_expected(current, config, session)
        return session["create"] << (
            st.task.new(self.create_task)(config.obj) >> expected
        )

    @property
    def create_task(self) -> Any:

        input_type = self.UP.input_type
        output_type = self.UP.output_type

        async def create_task(config: input_type) -> output_type:
            loop = asyncio.get_running_loop()

            check_resp, errs = await loop.run_in_executor(
                self.provider.thread_pool,
                self.provider.pulumi_provider.check,
                pylumi.URN(self.name),
                {},
                config,
            )
            check_resp = self.clean_check_resp(
                check_resp, self.UP.input_type, st.registry
            )
            if errs:
                raise PulumiValidationError(errs)
            resp = await loop.run_in_executor(
                self.provider.thread_pool,
                self.provider.pulumi_provider.create,
                pylumi.URN(self.name),
                check_resp,
                self.provider.operation_timeout,
            )
            return self.make_output(resp["Properties"], resp["ID"], output_type)

        return create_task

    async def modify(
        self,
        session: st.TaskSession,
        current: st.StateSnapshot,
        config: st.StateConfig,
    ) -> st.Object:
        expected = await self.get_expected(current, config, session)
        return session["modify"] << (
            st.task.new(self.modify_task)(current.obj, config.obj) >> expected
        )

    @property
    def modify_task(self) -> Any:

        input_type = self.UP.input_type
        output_type = self.UP.output_type

        async def modify_task(current: output_type, config: input_type) -> output_type:
            loop = asyncio.get_running_loop()
            current = current.copy()
            pulumi_id = current.pop(PULUMI_ID)

            check_resp, errs = await loop.run_in_executor(
                self.provider.thread_pool,
                self.provider.pulumi_provider.check,
                pylumi.URN(self.name, pulumi_id),
                current,
                config,
            )
            check_resp = self.clean_check_resp(
                check_resp, self.UP.input_type, st.registry
            )
            if errs:
                raise PulumiValidationError(errs)

            resp = await loop.run_in_executor(
                self.provider.thread_pool,
                self.provider.pulumi_provider.update,
                pylumi.URN(self.name, pulumi_id),
                pulumi_id,
                current,
                check_resp,
                self.provider.operation_timeout,
            )
            return self.make_output(resp["Properties"], resp["ID"], output_type)

        return modify_task

    async def delete(
        self, session: st.TaskSession, current: st.StateSnapshot
    ) -> st.Object:
        ref = session["delete"] << st.task.new(self.delete_task)(current.obj)
        return st.join(st.Object({}, st.EmptyType, session.ns.registry), ref)

    @property
    def delete_task(self) -> Any:

        null_type = self.DOWN.output_type
        output_type = self.UP.output_type

        async def delete_task(current: output_type) -> null_type:
            loop = asyncio.get_running_loop()
            current = current.copy()
            pulumi_id = current.pop(PULUMI_ID)

            resp = await loop.run_in_executor(
                self.provider.thread_pool,
                self.provider.pulumi_provider.delete,
                pylumi.URN(self.name, pulumi_id),
                pulumi_id,
                current,
                self.provider.operation_timeout,
            )
            return {}

        return delete_task

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name})"



import collections
import dataclasses as dc
import operator
from functools import reduce, partial
from typing import Type as PyType, Dict, Any, Union, Callable, Sequence, Optional

import statey as st
from statey.syms import types, utils, impl


# Default Plugin definitions
@dc.dataclass(frozen=True)
class HandleOptionalPlugin:
    """
    Handle an Optional[] annotation wrapper
    """

    @st.hookimpl
    def get_type(
        self, annotation: Any, registry: st.Registry, meta: Dict[str, Any]
    ) -> types.Type:
        inner = utils.extract_optional_annotation(annotation)
        if inner is None:
            return None
        meta["nullable"] = True
        return registry.get_type(inner, meta)


@dc.dataclass(frozen=True)
class ValuePredicatePlugin:
    """
    Simple predicate plugin that will may an annotation to a ValueType subclass (or any
    whos constructor is just the nullable argument)
    """

    predicate: Union[Callable[[Any], bool], PyType]
    type_cls: PyType[types.ValueType]

    @st.hookimpl(tryfirst=True)
    def get_type(
        self, annotation: Any, registry: st.Registry, meta: Dict[str, Any]
    ) -> types.Type:
        predicate = self.predicate
        if isinstance(self.predicate, type):
            predicate = lambda x: isinstance(x, type) and issubclass(x, self.predicate)

        if not predicate(annotation):
            return None

        meta = meta.copy()
        return self.type_cls(meta.pop("nullable", False), meta=meta)


@dc.dataclass(frozen=True)
class AnyPlugin:
    """
    Plugin that will always return AnyType. Should be added FIRST
    """

    @st.hookimpl
    def get_type(
        self, annotation: Any, registry: st.Registry, meta: Dict[str, Any]
    ) -> types.Type:
        return types.AnyType(meta=meta)


@dc.dataclass(frozen=True)
class ParseSequencePlugin:
    """
    Parse lists and sequences into ArrayTypes
    """

    array_type_cls: PyType[types.ArrayType] = types.ArrayType

    @st.hookimpl
    def get_type(
        self, annotation: Any, registry: st.Registry, meta: Dict[str, Any]
    ) -> types.Type:
        # Sort of hacky, will behave differently in different python versions (3.7 works)
        if not utils.is_sequence_annotation(annotation):
            return None

        inner = utils.extract_inner_annotation(annotation)
        # Optionals are subtypes of themselves I guess?
        if utils.extract_optional_annotation(annotation) is not None:
            return None
        element_type = registry.get_type(inner) if inner else st.Any
        meta = meta.copy()
        return self.array_type_cls(element_type, meta.pop("nullable", False), meta=meta)

    @st.hookimpl
    def infer_type(self, obj: Any, registry: "Registry") -> types.Type:
        if not isinstance(obj, (list, tuple)):
            return None

        element_types = []
        for item in obj:
            typ = registry.infer_type(item)
            if typ not in element_types:
                element_types.append(typ)

        if not element_types:
            return None
        if len(element_types) > 1:
            return None
        return types.ArrayType(element_types.pop(), False)


@dc.dataclass(frozen=True)
class ParseMappingPlugin:
    """
    Parse lists and sequences into ArrayTypes
    """

    map_type_cls: PyType[types.MapType] = types.MapType

    @st.hookimpl
    def get_type(
        self, annotation: Any, registry: st.Registry, meta: Dict[str, Any]
    ) -> types.Type:
        # Sort of hacky, will behave differently in different python versions (3.7 works)
        if not utils.is_mapping_annotation(annotation):
            return None

        inners = utils.extract_inner_annotations(annotation)
        # Optionals are subtypes of themselves I guess?
        if utils.extract_optional_annotation(annotation) is not None:
            return None
        if inners is not None and len(inners) != 2:
            return None
        key_annotation, value_annotation = inners or (None, None)
        key_type = registry.get_type(key_annotation) if key_annotation else st.Any
        value_type = registry.get_type(value_annotation) if value_annotation else st.Any
        meta = meta.copy()
        return self.map_type_cls(
            key_type, value_type, meta.pop("nullable", False), meta=meta
        )


@dc.dataclass(frozen=True)
class ParseDataClassPlugin:
    """
    Parse a specific dataclass into a StructType
    """

    dataclass_cls: PyType
    struct_type_cls: PyType[types.StructType] = types.StructType

    @st.hookimpl
    def get_type(
        self, annotation: Any, registry: st.Registry, meta: Dict[str, Any]
    ) -> types.Type:
        if annotation is not self.dataclass_cls or not dc.is_dataclass(annotation):
            return None
        fields = []
        for dc_field in utils.encodeable_dataclass_fields(annotation):
            field_annotation = dc_field.type
            syms_type = registry.get_type(field_annotation)
            syms_field = types.StructField(dc_field.name, syms_type)
            fields.append(syms_field)
        meta = meta.copy()
        meta.update(
            {
                "plugins": [
                    EncodeDataClassPlugin(self.dataclass_cls, self.struct_type_cls)
                ]
            }
        )
        instance = self.struct_type_cls(
            tuple(fields), meta.pop("nullable", False), meta=meta
        )
        return instance


@dc.dataclass(frozen=True)
class EncodeDataClassPlugin:
    """
    Parse a specific dataclass into a StructType
    """

    dataclass_cls: PyType
    struct_type_cls: PyType[types.StructType] = types.StructType

    @st.hookimpl
    def decode(self, value: Any) -> Any:
        return self.dataclass_cls(**value) if value is not None else None

    @st.hookimpl
    def encode(self, value: Any) -> Any:
        if not isinstance(value, self.dataclass_cls) or not dc.is_dataclass(value):
            return None
        return {
            field.name: getattr(value, field.name)
            for field in utils.encodeable_dataclass_fields(value)
        }


@dc.dataclass(frozen=True)
class LiteralPlugin:
    """
    Create Data literals from python objects whose types can be inferred directly
    """

    @st.hookimpl
    def get_object(self, value: Any, registry: st.Registry) -> "Object":
        if isinstance(value, st.Object):
            return None
        try:
            value_type = registry.infer_type(value)
        except st.exc.NoTypeFound:
            return None
        return st.Object(impl.Data(value, value_type), registry=registry)


@dc.dataclass(frozen=True)
class StructFromDictPlugin:
    """
    Attempts to construct a struct type given a dictionary input
    """

    @st.hookimpl
    def infer_type(self, obj: Any, registry: "Registry") -> types.Type:
        if not isinstance(obj, dict) or not all(isinstance(key, str) for key in obj):
            return None

        fields = []
        for name, value in obj.items():
            try:
                field_type = registry.infer_type(value)
            except st.exc.NoTypeFound:
                return None
            fields.append(st.Field(name, field_type))

        return types.StructType(fields, False)


@dc.dataclass(frozen=True)
class BasicObjectBehaviors:
    """
    Basic behavior for inferring types from Objects
    """

    @st.hookimpl
    def get_object(self, value: Any, registry: st.Registry) -> "Object":
        if isinstance(value, st.Object):
            return value
        return None

    @st.hookimpl
    def infer_type(self, obj: Any, registry: "Registry") -> types.Type:
        if isinstance(obj, st.Object):
            return obj._type
        return None

    @st.hookimpl
    def get_type(
        self, annotation: Any, registry: st.Registry, meta: Dict[str, Any]
    ) -> types.Type:
        if isinstance(annotation, types.Type):
            meta = meta.copy()
            type_as_nullable = annotation.with_nullable(
                meta.pop("nullable", annotation.nullable)
            )
            type_meta = type_as_nullable.meta.copy()
            type_meta.update(meta)
            return type_as_nullable.with_meta(type_meta)
        return None


DEFAULT_PLUGINS = [
    AnyPlugin(),
    HandleOptionalPlugin(),
    ParseSequencePlugin(types.ArrayType),
    ParseMappingPlugin(types.MapType),
    ValuePredicatePlugin(float, types.FloatType),
    ValuePredicatePlugin(int, types.IntegerType),
    ValuePredicatePlugin(list, partial(types.ArrayType, types.Any)),
    ValuePredicatePlugin(str, types.StringType),
    ValuePredicatePlugin(bool, types.BooleanType),
    ValuePredicatePlugin(range, partial(types.ArrayType, types.Integer)),
    ValuePredicatePlugin((lambda x: x is Any), types.AnyType),
    LiteralPlugin(),
    BasicObjectBehaviors(),
    StructFromDictPlugin(),
]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register default plugins
    """
    if registry is None:
        registry = st.registry

    for plugin in DEFAULT_PLUGINS:
        registry.register(plugin)
